# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
from abc import ABC
from dataclasses import is_dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Hashable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import torch

from megatron.energon.flavors.base_dataset import (
    BaseCoreDataset,
    PinMemoryMixin,
    Sample,
    SavableDataset,
)
from megatron.energon.flavors.crude import CrudeSample, CrudeWebdataset
from megatron.energon.task_encoder.cooking import Cooker
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers import (
    BatchDataset,
    BlendDataset,
    ConcatDataset,
    EpochizeDataset,
    GroupBatchDataset,
    LimitDataset,
    LogSampleDataset,
    MapDataset,
    ShuffleBufferDataset,
)

T = TypeVar("T")
V = TypeVar("V")
T_sample = TypeVar("T_sample")
T_encoded_sample = TypeVar("T_encoded_sample")
T_raw_batch = TypeVar("T_raw_batch")
T_batch = TypeVar("T_batch")


FeatureBatcher = Callable[[List[Any]], Any]


def generic_batch(batch: List[Any]) -> Any:
    """Based on the types/shapes of the batch: Will either pad and stack, or return as list.
    Recurses structures (dict, dataclass, namedtuple) and applies the same logic to each field."""
    if isinstance(batch[0], torch.Tensor):
        return batch_pad_stack(batch)
    elif isinstance(batch[0], dict):
        return {key: generic_batch([sample[key] for sample in batch]) for key in batch[0].keys()}
    elif is_dataclass(batch[0]):
        return type(batch[0])(
            **{
                field.name: generic_batch([getattr(sample, field.name) for sample in batch])
                for field in dataclasses.fields(batch[0])
            }
        )
    elif isinstance(batch[0], tuple) and hasattr(batch[0], "_fields"):
        # NamedTuple
        return type(batch[0])(
            **{
                field: generic_batch([getattr(sample, field) for sample in batch])
                for field in batch[0]._fields
            }
        )
    else:
        return batch_list(batch)


def batch_stack(batch: List[Any]) -> Any:
    """Stack a batch of tensors."""
    return torch.stack(batch, dim=0)


def batch_pad_stack(batch: List[Any]) -> Any:
    """Stack a batch of arbitrary-sized tensors padded with 0s."""
    max_size = [max(b.shape[dim] for b in batch) for dim in range(batch[0].ndim)]
    batch_tensor = batch[0].new_zeros((len(batch), *max_size))
    for i, b in enumerate(batch):
        batch_tensor[(i, *(slice(0, s) for s in b.shape))] = b
    # Pad all tensors to max_size
    return batch_tensor


def batch_list(batch: List[Any]) -> Any:
    """Stack a batch of tensors padded with 0s."""
    return batch


@dataclasses.dataclass
class Batch(PinMemoryMixin):
    """Base class for a batch dataclass. Provides a default implementation for pinning memory."""


class TaskEncoder(ABC, Generic[T_sample, T_encoded_sample, T_raw_batch, T_batch]):
    """
    Base class for task encoders.

    Task encoding follows these steps:
      0. Data comes from the dataset
      1. :meth:`megatron.energon.TaskEncoder.encode_sample` is called on each sample
      2. :meth:`megatron.energon.TaskEncoder.batch` is called on the list of encoded samples
      3. :meth:`megatron.energon.TaskEncoder.encode_batch` is called on the batch
      4. :meth:`megatron.energon.TaskEncoder.to_device` is called on the encoded batch
      5. resulting encoded batch is passed to the network
    """

    cookers: Sequence[Cooker] = ()

    def cook_crude_sample(self, sample: Union[T_sample, CrudeSample]) -> T_sample:
        if isinstance(sample, CrudeSample):
            for cooker in self.cookers:
                if cooker.is_match(sample):
                    return cooker.cook(sample)

            raise NotImplementedError(
                "You are using crude samples but not providing a way to cook them."
            )
        else:
            assert isinstance(sample, Sample), "Sample must be a complete Sample or a CrudeSample"
            return sample

    def encode_sample(
        self, sample: T_sample
    ) -> Union[T_encoded_sample, Generator[T_encoded_sample, None, None]]:
        """Encode a single sample. May raise :exc:`megatron.energon.SkipSample` to skip a sample.
        Alternatively, this can be a generator that yields (or ignores) new samples."""
        return sample

    def batch(self, samples: List[T_encoded_sample]) -> T_raw_batch:
        """Move a batch to a device. May raise :exc:`megatron.energon.SkipSample` to skip a batch."""
        return self._batch(samples, type(samples[0]))

    def batch_group_criterion(self, sample: T_encoded_sample) -> Hashable:
        """Return a group criterion for the sample. Default implementation does not group
        (effectively, it returns a single value (`None`), thus only one group is used). May raise
        :exc:`megatron.energon.SkipSample` to skip a batch."""
        return None

    def encode_batch(self, batch: T_raw_batch) -> Union[T_batch, Generator[T_batch, None, None]]:
        """Encode a batch of samples. May raise :exc:`megatron.energon.SkipSample` to skip a batch.
        Alternatively, this can be a generator that yields (or ignores) new batches."""
        return batch

    def _batch(
        self,
        samples: List[T_sample],
        result_type: Type[T_raw_batch],
        actions: Optional[Dict[str, FeatureBatcher]] = None,
        default_action: FeatureBatcher = generic_batch,
    ) -> T_raw_batch:
        """
        Batch a list of samples.

        Args:
            samples: The samples to batch
            result_type: Type of the result (might be dict, dataclass, or namedtuple)
            actions: For each field (=key), may specify a specific batcher
            default_action: The batcher to apply to all fields not in `action`

        Returns:
            The batched result
        """
        # Get dict of samples
        if isinstance(samples[0], dict):
            list_samples = {key: [sample[key] for sample in samples] for key in samples[0].keys()}
        elif is_dataclass(samples[0]):
            list_samples = {
                field.name: [getattr(sample, field.name) for sample in samples]
                for field in dataclasses.fields(samples[0])
            }
        elif isinstance(samples[0], tuple) and hasattr(samples[0], "_fields"):
            # NamedTuple
            list_samples = {
                field: [getattr(sample, field) for sample in samples]
                for field in samples[0]._fields
            }
        else:
            raise ValueError("Unrecognized sample type.")
        # Convert each field
        if actions is not None:
            list_samples = {
                key: default_action(value) if key not in actions else actions[key](value)
                for key, value in list_samples.items()
            }
        else:
            list_samples = {key: default_action(value) for key, value in list_samples.items()}
        # Construct result
        if issubclass(result_type, dict):
            return list_samples
        elif dataclasses.is_dataclass(result_type) or issubclass(result_type, tuple):
            # DataClass or NamedTuple
            return result_type(**list_samples)
        else:
            raise ValueError("Unrecognized result type.")

    def build_batch(
        self,
        dataset: SavableDataset[T_encoded_sample],
        batch_size: Optional[int],
        batch_drop_last: bool = False,
        worker_config: Optional[WorkerConfig] = None,
    ) -> SavableDataset[T_raw_batch]:
        """Applies the batcher to the dataset."""
        if (
            getattr(self.batch_group_criterion, "__func__", None)
            is not TaskEncoder.batch_group_criterion
        ):
            assert batch_size is not None, "batch_size must be set if batch_group_criterion is set"
            dataset = GroupBatchDataset(
                dataset,
                batch_size=batch_size,
                group_criterion=self.batch_group_criterion,
                batcher=self.batch,
                drop_last=batch_drop_last,
                worker_config=worker_config,
            )
        elif batch_size is not None:
            dataset = BatchDataset(
                dataset,
                batch_size=batch_size,
                batcher=self.batch,
                drop_last=batch_drop_last,
                worker_config=worker_config,
            )
            if getattr(self.encode_batch, "__func__", None) is not TaskEncoder.encode_batch:
                dataset = MapDataset(dataset, self.encode_batch)
        else:
            assert (
                getattr(self.encode_batch, "__func__", None) is TaskEncoder.encode_batch
            ), "batch_size is not set, but encode_batch is not the default."
        return dataset

    def build_cook_crude_sample(
        self,
        dataset: SavableDataset[Union[T_sample, dict]],
        worker_config: WorkerConfig,
    ) -> SavableDataset[T_sample]:
        """Applies the sample cooker to the dataset if we have cookers registered."""
        if (
            self.cookers
            or getattr(self.build_cook_crude_sample, "__func__", None)
            is not TaskEncoder.build_cook_crude_sample
        ):
            dataset = MapDataset(dataset, self.cook_crude_sample)
        return dataset

    def build_encode_sample(
        self,
        dataset: SavableDataset[T_sample],
        worker_config: WorkerConfig,
    ) -> SavableDataset[T_encoded_sample]:
        """Applies the sample encoder to the dataset."""
        if getattr(self.encode_sample, "__func__", None) is not TaskEncoder.encode_sample:
            dataset = MapDataset(dataset, self.encode_sample)
        return dataset

    def build_train_datasets(
        self,
        *,
        datasets: List[Tuple[BaseCoreDataset[T_sample], float]],
        worker_config: WorkerConfig,
        batch_size: int,
        batch_drop_last: bool,
        virtual_epoch_length: int = 0,
        shuffle_buffer_size: Optional[int] = None,
    ) -> SavableDataset[T_batch]:
        """Combines train datasets to a single dataset."""

        # Check if there's a CrudeWebdataset but no cookers
        for dataset, _ in datasets:
            if isinstance(dataset, CrudeWebdataset):
                assert self.cookers, "CrudeWebdataset found, but no cookers registered."

        if len(datasets) > 1:
            dataset = BlendDataset(
                *datasets,
                worker_config=worker_config,
            )
        elif len(datasets) == 1:
            dataset = datasets[0][0]
        else:
            raise ValueError("No datasets given.")
        if shuffle_buffer_size is not None and shuffle_buffer_size > 1:
            dataset = ShuffleBufferDataset(
                dataset,
                size=shuffle_buffer_size,
                worker_config=worker_config,
            )
        dataset = self.build_cook_crude_sample(dataset, worker_config=worker_config)
        dataset = self.build_encode_sample(dataset, worker_config=worker_config)
        dataset = self.build_batch(
            dataset,
            batch_size=batch_size,
            batch_drop_last=batch_drop_last,
            worker_config=worker_config,
        )
        if virtual_epoch_length > 0:
            dataset = EpochizeDataset(
                dataset,
                length=virtual_epoch_length,
                worker_config=worker_config,
            )
        if worker_config.should_log(level=1):
            dataset = LogSampleDataset(dataset, mode="train", worker_config=worker_config)
        return dataset

    def build_val_datasets(
        self,
        *,
        datasets: List[Tuple[BaseCoreDataset[T_sample], float]],
        worker_config: WorkerConfig,
        batch_size: int,
        batch_drop_last: bool,
        limit: Optional[int] = None,
    ) -> SavableDataset[T_batch]:
        """Combines train datasets to a single dataset."""

        # Check if there's a CrudeWebdataset but no cookers
        for dataset, _ in datasets:
            if isinstance(dataset, CrudeWebdataset):
                assert self.cookers, "CrudeWebdataset found, but no cookers registered."

        if len(datasets) > 1:
            dataset = ConcatDataset(*[dataset for dataset, _ in datasets])
        elif len(datasets) == 1:
            dataset = datasets[0][0]
        else:
            raise ValueError("No datasets given.")
        dataset = self.build_cook_crude_sample(dataset, worker_config=worker_config)
        dataset = self.build_encode_sample(dataset, worker_config=worker_config)
        dataset = self.build_batch(
            dataset,
            batch_size=batch_size,
            batch_drop_last=batch_drop_last,
            worker_config=worker_config,
        )
        if limit is not None and limit > 0:
            dataset = LimitDataset(
                dataset,
                length=limit,
                worker_config=worker_config,
                reset_after_epoch=True,
            )
        if worker_config.should_log(level=2):
            dataset = LogSampleDataset(dataset, mode="val", worker_config=worker_config)
        return dataset

    def get_current_batch_index(self) -> int:
        """Returns the current index for the next batch yielded from the current worker."""
        assert (
            WorkerConfig.active_worker_config is not None
        ), "The batch_index can only be fetched within the worker, and to be usable, you must use the get_(savable_)loader methods provided from the package."
        return WorkerConfig.active_worker_config.active_worker_sample_index


class DefaultTaskEncoder(
    TaskEncoder[T_sample, T_encoded_sample, T_raw_batch, T_batch],
    ABC,
    Generic[T_sample, T_encoded_sample, T_raw_batch, T_batch],
):
    """
    The default task encoder supports automagically mapping to target types.
    You may override any methods to customize the behavior. By default, `encode_sample` is the
    identity function, `batch` calls `_batch` with the type of the first sample, and `encode_batch`
    is also the identity function. If you set any of `encoded_sample_type`, 'raw_batch_type' or
    `batch_type`, the corresponding method return that type, where it automatically maps the fields
    (by name) to your new type.
    """

    _encoded_sample_type: Optional[Type[T_encoded_sample]]
    _raw_batch_type: Optional[Type[T_raw_batch]]
    _batch_type: Optional[Type[T_batch]]

    def __init__(
        self,
        *,
        encoded_sample_type: Optional[Type[T_encoded_sample]] = None,
        raw_batch_type: Optional[Type[T_raw_batch]] = None,
        batch_type: Optional[Type[T_batch]] = None,
    ):
        """
        Initialize the default task encoder.
        Types may be:
          * A `@dataclass` class: Return that typed dataclass. Field names must match the input
            fields.
          * A `NamedTuple` class: Return that typed namedtuple. Field names must match the input
            fields.
          * `dict`: Simply return the input as dict with field names as keys.

        Args:
            encoded_sample_type: Type of encoded samples (before batching)
            raw_batch_type: Type of the batched samples (after batching)
            batch_type: Type of the encoded batched samples
        """
        self._encoded_sample_type = encoded_sample_type
        self._raw_batch_type = raw_batch_type
        self._batch_type = batch_type

    def encode_sample(
        self, sample: T_sample
    ) -> Union[T_encoded_sample, Generator[T_encoded_sample, None, None]]:
        """Encode a single sample. The default implementation converts to the
        _encoded_sample_type."""
        if self._encoded_sample_type is None or isinstance(sample, self._encoded_sample_type):
            return sample
        if is_dataclass(sample):
            fields = {
                field.name: getattr(sample, field.name) for field in dataclasses.fields(sample)
            }
        elif isinstance(sample, tuple) and hasattr(sample, "_fields"):
            fields = {field: getattr(sample, field) for field in sample._fields}
        elif isinstance(sample, dict):
            fields = sample
        else:
            raise ValueError("Unrecognized sample type.")
        if issubclass(self._encoded_sample_type, dict):
            return fields
        elif dataclasses.is_dataclass(self._encoded_sample_type) or issubclass(
            self._encoded_sample_type, tuple
        ):
            # DataClass or NamedTuple
            return self._encoded_sample_type(**fields)
        else:
            raise ValueError("Unrecognized encoded sample type.")

    def batch(self, samples: List[T_encoded_sample]) -> T_raw_batch:
        """Batch a list of samples. The default implementation uses default batching to convert
        to _batch_type."""
        actions = None
        if isinstance(samples[0], Sample):
            actions = {
                "__subflavor__": lambda x: x,
                "__subflavors__": lambda x: x,
            }
        return self._batch(
            samples,
            type(samples[0]) if self._raw_batch_type is None else self._raw_batch_type,
            actions=actions,
        )

    def encode_batch(self, batch: T_raw_batch) -> Union[T_batch, Generator[T_batch, None, None]]:
        """Encode a batch of samples. The default implementation converts to the
        _encoded_batch_type."""
        if self._batch_type is None or self._raw_batch_type == self._batch_type:
            return batch
        if is_dataclass(batch):
            fields = {field.name: getattr(batch, field.name) for field in dataclasses.fields(batch)}
        elif isinstance(batch, tuple) and hasattr(batch, "_fields"):
            fields = {field: getattr(batch, field) for field in batch._fields}
        elif isinstance(batch, dict):
            fields = batch
        else:
            raise ValueError("Unrecognized sample type.")
        if issubclass(self._batch_type, dict):
            return fields
        elif dataclasses.is_dataclass(self._batch_type) or issubclass(self._batch_type, tuple):
            # DataClass or NamedTuple
            return self._batch_type(**fields)
        else:
            raise ValueError("Unrecognized encoded sample type.")


class AugmentTaskEncoder(
    TaskEncoder[T_sample, T_encoded_sample, T_raw_batch, T_batch],
    Generic[T_sample, T_encoded_sample, T_raw_batch, T_batch],
):
    """Augment a task encoder with additional functionality. By default, delegates everything to the
    original task encoder."""

    def __init__(self, task_encoder: TaskEncoder[T_sample, T_encoded_sample, T_raw_batch, T_batch]):
        """Initialize the augmenting task encoder.

        Args:
            task_encoder: The delegate task encoder. All calls will by default be forwarded to this.
        """
        self._task_encoder = task_encoder

    def encode_sample(self, sample: T_sample) -> T_encoded_sample:
        return self._task_encoder.encode_sample(sample)

    def batch(self, samples: List[T_encoded_sample]) -> T_raw_batch:
        return self._task_encoder.batch(samples)

    def batch_group_criterion(self, sample: T_encoded_sample) -> Hashable:
        return self._task_encoder.batch_group_criterion(sample)

    def encode_batch(self, batch_data: T_raw_batch) -> T_batch:
        return self._task_encoder.encode_batch(batch_data)



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Any, Dict, Generic, Iterator, List, Optional, Tuple, TypeVar, Union

from megatron.energon.flavors.base_dataset import MergedState, SavableDataset, State
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.base import BaseWrapperDataset

T_sample = TypeVar("T_sample")


@dataclass
class ConcatState(State):
    #: State of the inner datasets
    dataset_states: List[State]


@dataclass
class ConcatMergedState(MergedState):
    #: State of the inner datasets
    dataset_states: List[MergedState]


class ConcatDataset(BaseWrapperDataset[T_sample], Generic[T_sample]):
    """
    This dataset wrapper concatenates multiple iterable datasets together. The datasets must be
    finite, otherwise not all datasets can be sampled. This is only useful for validation / test
    datasets.
    """

    datasets: Tuple[SavableDataset[T_sample], ...]

    def __init__(self, *datasets: SavableDataset[T_sample]):
        """Construct a concatenated dataset."""
        super().__init__()
        self.datasets = datasets
        assert len(self) >= 0, "Datasets must be finite."

    def __len__(self):
        return sum(len(dataset) for dataset in self.datasets)

    def __iter__(self) -> Iterator[T_sample]:
        for ds_idx, dataset in enumerate(self.datasets):
            for sample in dataset:
                yield self._add_sample_restore_key(sample, ds_idx)

    def save_state(self) -> ConcatState:
        return ConcatState(dataset_states=[dataset.save_state() for dataset in self.datasets])

    def merge_states(self, states: List[ConcatState]) -> ConcatMergedState:
        assert all(s is None or isinstance(s, ConcatState) for s in states)
        assert all(s is None or len(s.dataset_states) == len(self.datasets) for s in states)
        return ConcatMergedState(
            dataset_states=[
                dataset.merge_states(
                    [None if s is None else s.dataset_states[ds_idx] for s in states]
                )
                for ds_idx, dataset in enumerate(self.datasets)
            ]
        )

    def restore_state(self, state: Optional[ConcatMergedState]) -> None:
        if state is None:
            for dataset in self.datasets:
                dataset.restore_state(None)
        else:
            assert isinstance(state, ConcatMergedState)
            assert len(self.datasets) == len(state.dataset_states)
            for dataset, dstate in zip(self.datasets, state.dataset_states):
                dataset.restore_state(dstate)

    def can_restore_sample(self) -> bool:
        return all(dataset.can_restore_sample() for dataset in self.datasets)

    def restore_sample(self, index: Tuple[Union[str, int], ...]) -> T_sample:
        ds_idx = index[0]
        assert isinstance(ds_idx, int)
        return self._add_sample_restore_key(self.datasets[ds_idx].restore_sample(index[1:]), ds_idx)

    def verify_worker_config(self, worker_config: WorkerConfig) -> None:
        super().verify_worker_config(worker_config)
        for dataset in self.datasets:
            dataset.verify_worker_config(worker_config)

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "datasets": [dataset.config() for dataset in self.datasets],
        }

    def __str__(self):
        return f"ConcatDataset(datasets={self.datasets})"



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import inspect
from typing import Any, Callable, Dict, Generator, Generic, Iterator, Tuple, TypeVar, Union

from megatron.energon.errors import SYSTEM_EXCEPTIONS, FatalSampleError
from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.wrappers._log_exception import log_exception
from megatron.energon.wrappers.base import BaseSingleWrapperDataset
from megatron.energon.wrappers.skip import SkipSample

T_sample = TypeVar("T_sample")
T_sample_out = TypeVar("T_sample_out")


class MapDataset(BaseSingleWrapperDataset[T_sample, T_sample_out], Generic[T_sample, T_sample_out]):
    """This dataset wrapper applies a custom function to transform each sample."""

    map_fn: Callable[[T_sample], Union[T_sample_out, Generator[T_sample_out, None, None]]]
    error_handler: Callable[[Exception, T_sample], None]
    stateless_map_fn: bool

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        map_fn: Callable[[T_sample], Union[T_sample_out, Generator[T_sample_out, None, None]]],
        error_handler: Callable[[Exception, T_sample], None] = log_exception,
        stateless_map_fn: bool = False,
    ):
        """Construct a MapDataset.

        If this should be savable, the map_fn must only return a sample, or a generator yielding
        0 or 1 sample per input sample. Otherwise this will be broken (see `IterMapDataset`).

        Args:
            dataset: The input dataset to wrap
            map_fn: The function to apply to each sample. May raise
                :exc:`megatron.energon.SkipSample` to skip a sample. Alternatively, may return a
                generator to yield multiple or no samples.
            error_handler: Handler for errors. Defaults to logging and ignoring the exception.
            stateless_map_fn: If true, the map_fn is deterministic and stateless
                (thus key for random access can propagate to inner dataset). Defaults to False.
        """
        super().__init__(dataset)
        self.map_fn = map_fn
        self.error_handler = error_handler
        self.stateless_map_fn = stateless_map_fn

    def __len__(self):
        return len(self.dataset)

    def __iter__(self) -> Iterator[T_sample_out]:
        for sample in self.dataset:
            try:
                mapped_sample = self.map_fn(sample)
                if isinstance(mapped_sample, Generator):
                    # In case of a generator, additionally store the index of the yielded samples
                    # per input sample
                    if self.can_restore_sample():
                        # If this is supposed to be restorable, map_fn must be a generator function
                        assert inspect.isgeneratorfunction(self.map_fn)
                        for idx, res_sample in enumerate(mapped_sample):
                            yield self._add_sample_restore_key(res_sample, idx, fail_otherwise=True)
                    else:
                        yield from mapped_sample
                else:
                    yield mapped_sample
            except SkipSample:
                pass
            except SYSTEM_EXCEPTIONS as e:
                raise FatalSampleError.from_sample(sample)
            except Exception as e:
                self.error_handler(e, sample)

    def can_restore_sample(self) -> bool:
        return self.stateless_map_fn and self.dataset.can_restore_sample()

    def restore_sample(self, index: Tuple[Union[str, int], ...]) -> T_sample_out:
        if self.stateless_map_fn:
            if inspect.isgeneratorfunction(self.map_fn):
                local_index = index[0]
                assert isinstance(local_index, int)
                index = index[1:]
            mapped_sample = self.map_fn(self.dataset.restore_sample(index))
            if isinstance(mapped_sample, Generator):
                assert inspect.isgeneratorfunction(self.map_fn)
                for idx, res_sample in enumerate(mapped_sample):
                    if idx == local_index:
                        return self._add_sample_restore_key(res_sample, idx, fail_otherwise=True)
                raise RuntimeError(
                    "Generator did not yield enough samples, but is marked stateless/deterministic."
                )
            else:
                return mapped_sample
        else:
            # Raise default error
            return super().__getitem__(index)

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
            "map_fn": self._function_config(self.map_fn),
        }

    def __str__(self):
        return f"MapDataset(map_fn={self.map_fn}, dataset={self.dataset})"



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import os
import re
from pathlib import Path as PathlibPath
from pathlib import PurePosixPath
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

import fsspec
import s3fs
from fsspec import AbstractFileSystem
from fsspec.implementations.local import LocalFileSystem

from megatron.energon.epathlib.rclone_config import ConfigEntry, read_rclone_config


class EPath:
    """EPath work much like pathlib.Path but does support local as well as remote file systems.
    To initialize a remote path, prepend the path with "rclone://". For example:

        EPath("rclone://myrclonename/myfile.txt")

    You will need to have your rclone configuration set up to access the remote file system.
    Currently only S3-based rclone remotes are supported.
    """

    internal_path: PurePosixPath  # The path without the protocol. Can also be in S3 for example
    fs: AbstractFileSystem
    protocol: str
    num_fs_path_parts: (
        int  # Number of parts in the internal_path that are part of the file system specification
    )
    s3_args: Optional[Dict[str, Any]]

    def __init__(
        self,
        initial_path: Union[str, "EPath", PathlibPath],
        config_override: Optional[Dict[str, ConfigEntry]] = None,
    ) -> None:
        self.s3_args = None

        if isinstance(initial_path, EPath):
            self.internal_path = initial_path.internal_path
            self.fs = initial_path.fs
            self.protocol = initial_path.protocol
            self.num_fs_path_parts = initial_path.num_fs_path_parts
            self.s3_args = initial_path.s3_args
        elif isinstance(initial_path, PathlibPath):
            self.internal_path = PurePosixPath(initial_path.absolute())
            self.fs = LocalFileSystem()
            self.protocol = "local"
            self.num_fs_path_parts = 0
        elif isinstance(initial_path, str):
            protocol, path = self._split_protocol(initial_path)
            if protocol is None:
                self.internal_path = PurePosixPath(path)
                self.fs = LocalFileSystem()
                self.protocol = "local"
                self.num_fs_path_parts = 0
            elif protocol == "rclone":
                if not path.startswith("/"):
                    # rclone paths are treated as absolute
                    # The first part of the path will be the rclone remote name
                    path = "/" + path
                self.internal_path = PurePosixPath(path)

                remote_name = self.internal_path.parts[1]
                self.fs, self.s3_args = self.create_s3fs_from_rclone_remote(
                    remote_name, config_override=config_override
                )
                self.protocol = "rclone"
                self.num_fs_path_parts = 2  # Root and remote name
            else:
                raise ValueError(f"Unknown protocol: {protocol}")

    @staticmethod
    def _split_protocol(path):
        regex = re.compile(r"^(?P<protocol>[a-z]+)://(?P<path>.+)$")
        m = regex.match(path)
        if m is None:
            return None, path
        return m.group("protocol"), m.group("path")

    @staticmethod
    def create_s3fs_from_rclone_remote(
        remote_name: str, config_override: Optional[Dict[str, ConfigEntry]] = None
    ) -> Tuple[s3fs.S3FileSystem, Dict[str, Any]]:
        """Given an rclone remote name (as in rclone.conf), create an S3FileSystem."""

        # Now read the rclone config file to get the endpoint URL and credentials
        if config_override is not None:
            rclone_config = config_override
        else:
            rclone_config = read_rclone_config()

        if remote_name not in rclone_config:
            raise ValueError(f"Unknown rclone remote: {remote_name}")

        rclone_entry = rclone_config[remote_name]

        if rclone_entry.type.lower() != "s3":
            raise NotImplementedError(
                f"Only S3 rclone remotes are supported. Remote {remote_name} is of type {rclone_entry.type}"
            )

        # If the endpoint does not specify a protocol, assume https://
        if rclone_entry.endpoint and not rclone_entry.endpoint.startswith("http"):
            s3fs_endpoint = "https://" + rclone_entry.endpoint
        else:
            s3fs_endpoint = rclone_entry.endpoint

        s3_args = dict(
            anon=False,
            key=rclone_entry.access_key_id,
            secret=rclone_entry.secret_access_key,
            endpoint_url=s3fs_endpoint,
            client_kwargs={
                "verify": False,
                "region_name": rclone_entry.region,
            },
        )

        return s3fs.S3FileSystem(**s3_args), s3_args  # type: ignore

    @staticmethod
    def prepare_forked_process():
        """Clear the cache of instances created by EPath and the async stuff.
        This is needed to avoid issues when forking a process that uses EPath.

        This static method needs to be called once in the child process after forking.
        It is independent of any EPath instances. For those, see the fork_guard() below.
        """

        fsspec.asyn.iothread[0] = None  # type: ignore
        fsspec.asyn.loop[0] = None  # type: ignore
        s3fs.S3FileSystem.clear_instance_cache()

    def fork_guard(self):
        """Check if this EPath instance is a forked copy with S3FS and re-create S3FS if needed.

        This method should be called before any operation that potentially uses the S3FS instance.
        """

        if isinstance(self.fs, s3fs.S3FileSystem) and self.fs._pid != os.getpid():
            assert self.s3_args is not None
            self.fs = s3fs.S3FileSystem(**self.s3_args)

    @property
    def _internal_fs_path(self) -> str:
        """Return the part of path the path that specifies FS properties like rclone rmeote name."""
        return str(PurePosixPath("/", *self.internal_path.parts[: self.num_fs_path_parts]))

    @property
    def _internal_nonfs_path(self) -> str:
        """Return the path as used inside the file system, without the protocol and fs part."""
        return str(PurePosixPath("/", *self.internal_path.parts[self.num_fs_path_parts :]))

    def open(self, mode="r", block_size=None):
        self.fork_guard()
        assert self.is_absolute()
        return self.fs.open(self._internal_nonfs_path, mode, block_size=block_size)

    def read_text(self):
        with self.open() as f:
            return f.read()

    @property
    def name(self):
        return self.internal_path.name

    @property
    def parent(self):
        new_path = EPath(self)
        new_path.internal_path = self.internal_path.parent
        return new_path

    @property
    def url(self):
        assert self.is_absolute(), "Can only call url on absolute EPath"
        if self.protocol == "rclone":
            int_path_str = str(self.internal_path)
            if int_path_str.startswith("/"):
                # Strip leading / for display purposes
                int_path_str = int_path_str[1:]
            return f"{self.protocol}://{int_path_str}"
        else:
            return str(self.internal_path)

    @property
    def relpath(self):
        assert not self.is_absolute(), "Can only call relpath on relative EPath"
        return str(self.internal_path)

    def is_absolute(self):
        return self.internal_path.is_absolute()

    def absolute(self) -> "EPath":
        if self.protocol == "local":
            if self.is_absolute():
                return self
            return EPath(PathlibPath(self.internal_path).absolute())
        else:
            assert self.is_absolute(), "Remote paths are always absolute"
            return self

    def is_dir(self):
        self.fork_guard()
        assert self.is_absolute()
        return self.fs.isdir(self._internal_nonfs_path)

    def is_file(self):
        self.fork_guard()
        assert self.is_absolute()
        return self.fs.isfile(self._internal_nonfs_path)

    def mkdir(self, exist_ok: bool = True, parents: bool = False):
        self.fork_guard()
        assert self.is_absolute()
        if parents:
            return self.fs.makedirs(self._internal_nonfs_path, exist_ok=exist_ok)
        else:
            try:
                return self.fs.mkdir(self._internal_nonfs_path)
            except FileExistsError:
                if exist_ok:
                    pass
                else:
                    raise

    def glob(self, pattern) -> Generator["EPath", None, None]:
        self.fork_guard()
        assert self.is_absolute()

        search_path_pattern = (self / pattern)._internal_nonfs_path

        if self.protocol != "local" and search_path_pattern.startswith("/"):
            # For some reason s3fs glob does not like leading /
            search_path_pattern = search_path_pattern[1:]

        for path in self.fs.glob(search_path_pattern):
            assert isinstance(path, str)

            new_path = EPath(self)  # Copy

            if self.protocol == "local":
                assert path.startswith("/"), "Local FS glob should return absolute paths"
                new_path.internal_path = PurePosixPath(path)
            else:
                new_path.internal_path = self._internal_fs_path / PurePosixPath(path)

            yield new_path

    def size(self):
        self.fork_guard()
        assert self.is_absolute()
        return self.fs.size(self._internal_nonfs_path)

    def with_suffix(self, suffix):
        new_path = EPath(self)
        new_path.internal_path = self.internal_path.with_suffix(suffix)
        return new_path

    def move(self, target: "EPath"):
        self.fork_guard()
        assert self.is_absolute()
        assert target.is_absolute()

        assert self.fs == target.fs, "Can only move within same FS"
        assert self.protocol == target.protocol, "Can only move within same FS"
        assert self._internal_fs_path == target._internal_fs_path, "Can only move within same FS"

        self.fs.mv(self._internal_nonfs_path, target._internal_nonfs_path)

    def unlink(self):
        self.fork_guard()
        assert self.is_absolute()
        return self.fs.rm(self._internal_nonfs_path)

    def relative_to(self, other: "EPath") -> str:
        self.fork_guard()
        assert self.is_absolute()
        assert other.is_absolute()

        assert self.fs == other.fs, "Can only use relative_to within same FS"
        assert self.protocol == other.protocol, "Can only use relative_to within same FS"
        assert (
            self._internal_fs_path == other._internal_fs_path
        ), "Can only use relative_to within same FS"

        return str(self.internal_path.relative_to(other.internal_path))

    def __truediv__(self, other: Union[str, "EPath"]):
        assert self.is_absolute()

        other = EPath(other) if isinstance(other, str) else other
        if other.is_absolute():
            # If the second path is absolute, just return it as is
            return other

        new_path = EPath(self)
        new_path.internal_path = self.internal_path / (
            other if isinstance(other, str) else other.internal_path
        )
        return new_path

    def __rtruediv__(self, other: "EPath"):
        assert other.is_absolute()
        # assert not self.is_absolute()

        if self.is_absolute():
            # If second path (self) is absolute, just return it as is
            return self

        new_path = EPath(self)
        new_path.internal_path = other.internal_path / self.internal_path
        return new_path

    def __lt__(self, other: "EPath"):
        self.fork_guard()
        assert self.fs == other.fs, "Cannot compare paths from different file systems"
        assert self.protocol == other.protocol, "Cannot compare paths from different file systems"
        assert (
            self._internal_fs_path == other._internal_fs_path
        ), "Cannot compare paths from different file systems"

        return self.internal_path < other.internal_path

    def __str__(self):
        if self.is_absolute():
            return str(self.url)
        else:
            return str(self.relpath)

    def __repr__(self) -> str:
        return f"EPath({str(self)})"

    def __hash__(self):
        return hash((self.internal_path, self.protocol))

    def __eq__(self, other):
        return self.internal_path == other.internal_path and self.protocol == other.protocol



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import List, Union

import torch

from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.base_webdataset import DefaultDecoderWebdataset


@dataclass
class InterleavedSample(Sample):
    """Sample type for interleaved media such as text with images."""

    #: The interleaved media (either torch.tensor for an image, or str for text)
    sequence: List[Union[torch.Tensor, str]]


class InterleavedWebdataset(DefaultDecoderWebdataset[InterleavedSample]):
    __sample_type__ = InterleavedSample



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Hashable,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from megatron.energon.errors import SYSTEM_EXCEPTIONS, FatalSampleError
from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.worker import WorkerConfig, global_worker_config
from megatron.energon.wrappers._log_exception import log_exception
from megatron.energon.wrappers.base import (
    BaseSingleWrapperDataset,
    BaseSingleWrapperMergedState,
    BaseSingleWrapperState,
)
from megatron.energon.wrappers.skip import SkipSample

T_batch = TypeVar("T_batch", covariant=True)
T_batch_sample = TypeVar("T_batch_sample", covariant=True)


@dataclass
class GroupBatchState(BaseSingleWrapperState):
    batches: Any


@dataclass
class GroupBatchMergedState(BaseSingleWrapperMergedState):
    batches: List[Any]


class GroupBatchDataset(
    BaseSingleWrapperDataset[T_batch_sample, T_batch], Generic[T_batch_sample, T_batch]
):
    """This dataset wrapper transforms a dataset of samples into a dataset of batches, grouped
    by some criterion. This will not have a correct length, as it depends on the grouping.
    An example use case is: Image-Text samples, which are to be grouped by the image size into three
    size categories (e.g. 128x128, 256x256, 512x512) for efficient augmentation and batching.
    """

    dataset: SavableDataset[T_batch_sample]
    batch_size: int
    group_criterion: Callable[[T_batch_sample], Hashable]
    batcher: Callable[[List[T_batch_sample]], T_batch]
    drop_last: bool
    error_handler: Callable[[Exception, List[T_batch_sample]], None]
    worker_config: Optional[WorkerConfig]
    n_groups: int

    _state_batches: List[Optional[Dict[Hashable, List[T_batch_sample]]]]

    def __init__(
        self,
        dataset: SavableDataset[T_batch_sample],
        batch_size: int,
        group_criterion: Callable[[T_batch_sample], Hashable],
        batcher: Callable[[List[T_batch_sample]], T_batch],
        drop_last: bool = False,
        error_handler: Callable[[Exception, List[T_batch_sample]], None] = log_exception,
        worker_config: Optional[WorkerConfig] = None,
        n_groups: int = 1,
    ):
        """Construct a GroupBatchDataset.

        Args:
            dataset: The input dataset to wrap
            batch_size: The desired batch size. The last batch may be smaller.
            group_criterion: Function which determines the group of a sample.
            batcher: Function which combines separate samples into a single object. May raise
                :exc:`megatron.energon.SkipSample` to skip a sample.
            drop_last: If True, the last batch is dropped if it is smaller than the batch size.
            error_handler: Handler for errors. Defaults to logging and ignoring the exception.
            worker_config: Configuration for the workers. Defaults to `global_worker_config`.
            n_groups: Number of different groups. If not set properly, `len` might be less than the
                actual number of samples yielded.
        """
        super().__init__(dataset)
        self.batch_size = batch_size
        self.group_criterion = group_criterion
        self.batcher = batcher
        self.drop_last = drop_last
        self.error_handler = error_handler
        self.worker_config = worker_config or global_worker_config
        self.n_groups = n_groups
        self._state_batches = [{} for _ in range(max(self.worker_config.num_workers, 1))]

    def __len__(self):
        # This is only an approximation, since we don't know how many groups there are.
        # Should always overestimate (might yield less than actual number of batches).
        n_samples = len(self.dataset)
        worker_groups = self.worker_config.num_workers * self.n_groups
        n_samples_per_worker_floor = n_samples // worker_groups
        remaining_n_sample_workers = n_samples % worker_groups
        n_batches_per_worker_floor = n_samples_per_worker_floor // self.batch_size
        if n_samples_per_worker_floor % self.batch_size != 0 and not self.drop_last:
            n_batches_per_worker_floor += 1
        # Correct number of batches for the workers which yield 1 more sample (to balance)
        n_batches_per_worker_ceil = (n_samples_per_worker_floor + 1) // self.batch_size
        if n_batches_per_worker_ceil % self.batch_size != 0 and not self.drop_last:
            n_batches_per_worker_ceil += 1

        return (
            n_batches_per_worker_floor * (worker_groups - remaining_n_sample_workers)
            + n_batches_per_worker_ceil * remaining_n_sample_workers
        )

    def __iter__(self) -> Iterator[T_batch]:
        batches: Dict[Hashable, List[T_batch_sample]]
        worker_idx = self.worker_config.rank_worker_id()
        # Cleanup other states
        for i in range(self.worker_config.num_workers):
            if i != worker_idx:
                self._state_batches[i] = None
        if self._state_batches[worker_idx] is None:
            self._state_batches[worker_idx] = {}
        batches = self._state_batches[worker_idx]
        for sample in self.dataset:
            try:
                group = self.group_criterion(sample)
            except SkipSample:
                continue
            except SYSTEM_EXCEPTIONS:
                raise FatalSampleError.from_sample(sample)
            except Exception as e:
                self.error_handler(e, [sample])
                continue
            batch_group = batches.get(group)
            if batch_group is None:
                batches[group] = batch_group = []
            batch_group.append(sample)
            if len(batch_group) == self.batch_size:
                try:
                    yield self.batcher(batch_group)
                except SkipSample:
                    pass
                except SYSTEM_EXCEPTIONS:
                    raise FatalSampleError.from_sample(batch_group)
                except Exception as e:
                    self.error_handler(e, batch_group)
                del batches[group]
        if not self.drop_last:
            for batch in batches.values():
                try:
                    yield self.batcher(batch)
                except SkipSample:
                    pass
                except SYSTEM_EXCEPTIONS:
                    raise FatalSampleError.from_sample(batch)
                except Exception as e:
                    self.error_handler(e, batch)
        self._state_batches[worker_idx] = None

    def save_state(self) -> GroupBatchState:
        return GroupBatchState.extend(
            super().save_state(),
            batches=dict(self._state_batches[self.worker_config.rank_worker_id()]),
        )

    def merge_states(self, states: List[GroupBatchState]) -> GroupBatchMergedState:
        assert all(s is None or isinstance(s, GroupBatchState) for s in states)
        return GroupBatchMergedState.extend(
            super().merge_states(states),
            batches=[{} if s is None else s.batches for s in states],
        )

    def restore_state(self, state: GroupBatchMergedState) -> None:
        super().restore_state(state)
        if state is None:
            self._state_batches = [{} for _ in range(max(self.worker_config.num_workers, 1))]
        else:
            assert isinstance(state, GroupBatchMergedState)
            self._state_batches = state.batches

    def can_restore_sample(self) -> bool:
        return False

    def restore_sample(self, index: Tuple[Union[str, int], ...]) -> T_batch:
        # TODO: We'd need to store multiple indices to restore a batch
        # Also, returned elements don't support __restore_key__. Would need extension.
        raise NotImplementedError("GroupBatchDataset does not support random access.")

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
            "batch_size": self.batch_size,
            "group_criterion": self._function_config(self.group_criterion),
            "batcher": self._function_config(self.batcher),
            "drop_last": self.drop_last,
            "error_handler": self._function_config(self.error_handler),
            "worker_config": self.worker_config.config(),
            "n_groups": self.n_groups,
        }

    def __str__(self):
        return f"GroupBatchDataset(batch_size={self.batch_size}, group_criterion={self.group_criterion}, batcher={self.batcher}, drop_last={self.drop_last}, error_handler={self.error_handler}, worker_config={self.worker_config}, n_groups={self.n_groups}, dataset={self.dataset})"



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterator,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from torch.utils.data import IterableDataset

from megatron.energon.errors import SYSTEM_EXCEPTIONS, FatalSampleError
from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.wrappers._log_exception import log_exception
from megatron.energon.wrappers.base import BaseSingleWrapperDataset

T_sample = TypeVar("T_sample")
T_sample_out = TypeVar("T_sample_out")


class _ExcSampler:
    last_sample: Optional[T_sample] = None
    dataset: IterableDataset[T_sample]

    def __init__(self, dataset: IterableDataset[T_sample]):
        self.dataset = dataset

    def __iter__(self) -> Iterator[T_sample]:
        for sample in self.dataset:
            self.last_sample = sample
            yield sample


class IterMapDataset(
    BaseSingleWrapperDataset[T_sample, T_sample_out], Generic[T_sample, T_sample_out]
):
    """This dataset wrapper applies a custom function to transform the stream of samples and yield
    a new stream of samples.
    If used in a savable dataset context, it is critical, that `iter_map_fn` is either stateless,
    or that the state of the `iter_map_fn` is saved and restored externally.
    """

    iter_map_fn: Callable[[Iterator[T_sample]], Iterator[T_sample_out]]
    len_map_fn: Callable[[int], int]
    error_handler: Callable[[Exception, Optional[T_sample]], None]
    stateless_iter_fn: bool

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        iter_map_fn: Callable[[Iterator[T_sample]], Iterator[T_sample_out]],
        len_map_fn: Callable[[int], int] = lambda x: x,
        error_handler: Callable[[Exception, Optional[T_sample]], None] = log_exception,
        stateless_iter_fn: bool = False,
    ):
        """Construct a IterMapDataset.
        For saving and restoring samples, the iter_map_fn must only yield 0 or 1 sample per
        iterated sample.

        TODO: Implement saving/restoring for arbitrary number of yielded samples.
        Problem is, that this would require the inner dataset to be restored on the previous sample,
        such that this dataset can skip the already yielded samples.

        Args:
            dataset: The input dataset to wrap
            iter_map_fn: The function to apply to the stream of samples. Returns a new stream of
                samples. If savability should be preserved, this function should be stateless.
            len_map_fn: The function to apply to the length of the dataset. Returns the new
                (approximate) length of the resulting stream of samples based on the original
                length.
            error_handler: Handler for errors. Defaults to logging and ignoring the exception.
            stateless_iter_fn: If true, assume the iter_map_fn is deterministic and stateless
                (it does not aggregate samples (thus key for random access can propagate to inner
                dataset), yielding zero or multiple samples per fetched sample is fine).
                Defaults to False.
        """
        super().__init__(dataset)
        self.iter_map_fn = iter_map_fn
        self.len_map_fn = len_map_fn
        self.error_handler = error_handler
        self.stateless_iter_fn = stateless_iter_fn

    def __len__(self):
        return self.len_map_fn(len(self.dataset))

    def __iter__(self) -> Iterator[T_sample_out]:
        exc_sampler = _ExcSampler(self.dataset)
        if self.can_restore_sample():
            # The iter_map_fn is stateless. Thus we need to know which inner sample created the
            # outer sample, and the relative outer sample index, so we can restore it.

            # This is the sample index within the currently yielded sample
            iter_idx = 0

            def reset_idx_iter() -> Generator[T_sample, None, None]:
                # Resets the inner sample index
                nonlocal iter_idx
                for entry in exc_sampler:
                    iter_idx = 0
                    yield entry

            ds_iter = iter(reset_idx_iter())
            # While True will break when the inner dataset is exhausted, but continue on exception
            while True:
                iter_idx = 0
                try:
                    for res_sample in self.iter_map_fn(ds_iter):
                        yield self._add_sample_restore_key(
                            res_sample, iter_idx, fail_otherwise=True
                        )
                        iter_idx += 1
                except SYSTEM_EXCEPTIONS:
                    raise FatalSampleError.from_sample(exc_sampler.last_sample)
                except Exception as e:
                    self.error_handler(e, exc_sampler.last_sample)
                else:
                    break
        else:
            ds_iter = iter(exc_sampler)
            # While True will break when the inner dataset is exhausted, but continue on exception
            while True:
                try:
                    yield from self.iter_map_fn(ds_iter)
                except SYSTEM_EXCEPTIONS:
                    raise FatalSampleError.from_sample(exc_sampler.last_sample)
                except Exception as e:
                    self.error_handler(e, exc_sampler.last_sample)
                else:
                    break

    def can_restore_sample(self) -> bool:
        return self.stateless_iter_fn and self.dataset.can_restore_sample()

    def restore_sample(self, index: Tuple[Union[str, int], ...]) -> T_sample:
        assert self.stateless_iter_fn
        iter_idx = index[0]
        assert isinstance(iter_idx, int)
        inner_sample = self.dataset.restore_sample(index[1:])
        # Skip inner yielded samples to get the correct sample
        for idx, sample in enumerate(self.iter_map_fn(iter([inner_sample]))):
            if idx == iter_idx:
                return self._add_sample_restore_key(sample, idx, fail_otherwise=True)
        raise RuntimeError(
            "Generator did not yield enough samples, but is marked stateless/deterministic."
        )

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
            "iter_map_fn": self._function_config(self.iter_map_fn),
            "len_map_fn": self._function_config(self.len_map_fn),
            "error_handler": self._function_config(self.error_handler),
        }

    def __str__(self):
        return f"IterMapDataset(iter_map_fn={self.iter_map_fn}, dataset={self.dataset})"



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Hashable, List, Optional, TypeVar

import torch
import torch.distributed
import torch.utils.data

from megatron.energon.flavors.base_dataset import State
from megatron.energon.worker import WorkerConfig

T = TypeVar("T")


@dataclass
class WorkerRngState(State):
    """The state of a worker random generator."""

    rng: Optional[bytes]

    def __repr__(self):
        if self.rng is None:
            return "WorkerRngState(rng=None)"
        else:
            inner = None if self.rng is None else self.rng[:3] + b"..."
            return f"WorkerRngState(rng={inner!r})"


@dataclass
class WorkerRngMergedState(State):
    """The state of a worker random generator."""

    rng: List[Optional[bytes]]

    def __repr__(self):
        if self.rng is None:
            return "WorkerRngMergedState(rng=None)"
        else:
            inner = [None if r is None else r[:3] + b"..." for r in self.rng]
            return f"WorkerRngMergedState(rng={inner!r})"


class WorkerRng:
    """Helper class for getting a worker random generator, which is still in itself deterministic.
    If not in a worker, uses the global random generator's seed to initialize a new rng."""

    worker_config: WorkerConfig

    _rng: Optional[torch.Generator] = None

    _restore_states: Optional[List[Optional[bytes]]] = None

    def __init__(self, worker_config: WorkerConfig):
        self.worker_config = worker_config

    @property
    def rng(self) -> torch.Generator:
        if self._rng is None or self._restore_states is not None:
            self.worker_config.assert_worker()
            self._rng = torch.Generator()
            worker_idx = self.worker_config.rank_worker_id()
            if self._restore_states is not None and self._restore_states[worker_idx] is not None:
                self._rng.set_state(
                    torch.frombuffer(
                        bytearray(self._restore_states[worker_idx]),
                        dtype=torch.uint8,
                    ).clone()
                )
            else:
                # Restore to initial state (either due to zero sized states, or just initial state)
                self._rng.manual_seed(self.worker_config.worker_seed())
            self._restore_states = None
        return self._rng

    def randbelow(self, n: int) -> int:
        return torch.randint(0, n, (), generator=self.rng).item()

    def choice_idx(self, probs: torch.Tensor) -> int:
        if len(probs) == 1:
            return 0
        else:
            return torch.multinomial(probs, 1, replacement=True, generator=self.rng).item()

    def choice(self, l: List[T], probs: Optional[torch.Tensor] = None) -> T:
        if probs is None:
            return l[self.randbelow(len(l))]
        assert len(l) == len(probs)
        return l[self.choice_idx(probs)]

    def shuffle(self, l: List[T]) -> List[T]:
        """Returns a new list with shuffled entries"""
        p = torch.randperm(len(l), generator=self.rng)
        return [l[p[i]] for i in range(len(l))]

    def rand_pop(self, l: List[T]) -> T:
        return l.pop(self.randbelow(len(l)))

    def save_state(self) -> WorkerRngState:
        if self.rng is None:
            return WorkerRngState(rng=None)
        return WorkerRngState(rng=bytes(self.rng.get_state().tolist()))

    def merge_states(self, states: List[Optional[WorkerRngState]]) -> WorkerRngMergedState:
        assert all(s is None or isinstance(s, WorkerRngState) for s in states)
        return WorkerRngMergedState(rng=[None if s is None else s.rng for s in states])

    def restore_state(self, state: Optional[WorkerRngMergedState]) -> None:
        if state is None:
            self._restore_states = None
        else:
            assert isinstance(state, WorkerRngMergedState)
            self._restore_states = state.rng



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from megatron.energon.task_encoder.base import (
    AugmentTaskEncoder,
    Batch,
    DefaultTaskEncoder,
    TaskEncoder,
    batch_list,
    batch_pad_stack,
    batch_stack,
    generic_batch,
)
from megatron.energon.task_encoder.loader import (
    get_train_dataset,
    get_val_dataset,
    get_val_datasets,
)



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass

import torch

from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.base_webdataset import DefaultDecoderWebdataset


@dataclass
class ImageSample(Sample):
    """Sample type for an image, e.g. for image reconstruction."""

    #: The input image tensor in the shape (C, H, W)
    image: torch.Tensor


class ImageWebdataset(DefaultDecoderWebdataset[ImageSample]):
    __sample_type__ = ImageSample



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Any, Dict, Generic, Iterator, List, Optional, TypeVar

from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.worker import WorkerConfig, global_worker_config
from megatron.energon.wrappers.base import (
    BaseSingleWrapperDataset,
    BaseSingleWrapperMergedState,
    BaseSingleWrapperState,
)

T_sample = TypeVar("T_sample")


@dataclass
class LimitState(BaseSingleWrapperState):
    offset: int


@dataclass
class LimitMergedState(BaseSingleWrapperMergedState):
    offset: List[int]


class LimitDataset(BaseSingleWrapperDataset[T_sample, T_sample], Generic[T_sample]):
    """Limits the length of the dataset."""

    dataset: SavableDataset[T_sample]
    length: int
    worker_config: WorkerConfig

    _current_offset: List[int]

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        length: int,
        reset_after_epoch: bool = False,
        worker_config: Optional[WorkerConfig] = None,
    ):
        """
        Limits the length of the dataset.

        Args:
            dataset: The dataset to limit
            length: The length to limit to
            reset_after_epoch: If true, reset the underlying dataset after one epoch.
            worker_config: Configuration for the workers. Defaults to `global_worker_config`.
        """
        super().__init__(dataset)
        self.length = length
        self.reset_after_epoch = reset_after_epoch
        self.worker_config = worker_config or global_worker_config
        self._current_offset = [0] * max(self.worker_config.num_workers, 1)

    def __len__(self) -> int:
        return min(self.length, len(self.dataset))

    def __iter__(self) -> Iterator[T_sample]:
        worker_id = self.worker_config.rank_worker_id()

        # Compute the local limit for this worker, i.e. all worker's limits sum up to the total
        if self.worker_config.num_workers <= 1:
            local_limit = self.length
        else:
            local_limit = self.length // self.worker_config.num_workers
            if worker_id < self.length % self.worker_config.num_workers:
                local_limit += 1

        if self.worker_config.should_log(level=2):
            self.worker_config.worker_log(
                {
                    "t": "LimitDataset.start",
                    "r": self.worker_config.rank,
                    "w": worker_id,
                    "offset": self._current_offset[worker_id],
                    "local_limit": local_limit,
                    "limit": self.length,
                }
            )

        offset_range = list(
            range(self._current_offset[self.worker_config.rank_worker_id()], local_limit)
        )
        # Only iterate self.dataset if there are samples to iterate
        if len(offset_range) > 0:
            for sample, offset in zip(
                self.dataset,
                offset_range,
            ):
                self._current_offset[worker_id] = offset + 1
                yield sample

        if self.worker_config.should_log(level=2):
            self.worker_config.worker_log(
                {
                    "t": "LimitDataset.done",
                    "r": self.worker_config.rank,
                    "w": worker_id,
                    "offset": self._current_offset[worker_id],
                    "local_limit": local_limit,
                    "limit": self.length,
                }
            )

        # Reset the inner dataset
        self.dataset.restore_state(None)
        self._current_offset = [0] * max(self.worker_config.num_workers, 1)
        if self.reset_after_epoch:
            self.dataset.restore_state(None)

    def save_state(self) -> LimitState:
        return LimitState.extend(
            super().save_state(),
            offset=self._current_offset[self.worker_config.rank_worker_id()],
        )

    def merge_states(self, states: List[LimitState]) -> LimitMergedState:
        assert all(s is None or isinstance(s, LimitState) for s in states)
        return LimitMergedState.extend(
            super().merge_states(states),
            offset=[0 if s is None else s.offset for s in states],
        )

    def restore_state(self, state: Optional[LimitMergedState]) -> None:
        super().restore_state(state)
        if state is None:
            self._current_offset = [0] * max(self.worker_config.num_workers, 1)
        else:
            assert isinstance(state, LimitMergedState)
            self._current_offset = state.offset

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
            "length": self.length,
            "reset_after_epoch": self.reset_after_epoch,
            "worker_config": self.worker_config.config(),
        }

    def __str__(self):
        return f"LimitDataset(length={self.length}, dataset={self.dataset})"



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import gc
import multiprocessing
import queue
import random
import threading
import time
import traceback
import weakref
from dataclasses import dataclass
from functools import partial
from typing import Any, ClassVar, Dict, Generic, Iterator, List, Optional, Tuple, TypeVar, Union

import numpy as np
import torch
import torch.distributed
import torch.multiprocessing
from torch.utils.data import DataLoader, IterableDataset

from megatron.energon.flavors.base_dataset import MergedState, SavableDataset, State
from megatron.energon.worker import WorkerConfig, global_worker_config
from megatron.energon.wrappers.gc_dataset import GcDataset, gc_init_worker
from megatron.energon.wrappers.log_sample_dataset import default_get_keys

T = TypeVar("T")


def _init_worker(seed_per_worker: List[int], worker_id: int):
    """Initializes the the worker process.

    Sets the random seeds and prepare EPath for the forked worker process.
    """

    import random

    import numpy

    from megatron.energon.epathlib import EPath

    gc_init_worker(worker_id)

    worker_seed = seed_per_worker[worker_id]
    
    torch.manual_seed(worker_seed)
    numpy.random.seed(worker_seed)
    random.seed(worker_seed)

    EPath.prepare_forked_process()


class SimpleSavableDatasetWrapper(SavableDataset[Tuple[int, int, T]], Generic[T]):
    """Wrapper for non-multiprocessing savable datasets. Restarts the inner dataset. This class is
    not intended to be used directly."""

    dataset: SavableDataset[T]
    worker_config: WorkerConfig
    _state_restored: bool = False

    def __init__(self, dataset: SavableDataset[T], worker_config: WorkerConfig):
        self.dataset = dataset
        self.worker_config = worker_config

    def __len__(self):
        return len(self.dataset)

    def __iter__(self):
        self._state_restored = True
        sample_index = 0
        worker_id = self.worker_config.rank_worker_id()
        while self._state_restored:
            self._state_restored = False
            self.worker_config.worker_activate(sample_index)
            try:
                for src_data in self.dataset:
                    self.worker_config.worker_deactivate()
                    yield worker_id, sample_index, src_data
                    sample_index += 1
                    if self._state_restored:
                        # Restart iterator after restore
                        break
                    self.worker_config.worker_activate(sample_index)
            finally:
                self.worker_config.worker_deactivate()

    def save_state(self) -> State:
        return self.dataset.save_state()

    def merge_states(self, states: List[State]) -> MergedState:
        return self.dataset.merge_states(states)

    def restore_state(self, state: MergedState):
        self.dataset.restore_state(state)
        self._state_restored = True

    def config(self) -> Dict[str, Any]:
        return self.dataset.config()

    def __str__(self):
        return f"SimpleSavableDatasetWrapper(dataset={self.dataset})"


@dataclass
class SavableDatasetState(State):
    """State of the dataset wrapper. It stores the global random states and the index of the next
    sample to be returned from the dataset. This class is not intended to be used directly, but by
    :class:`megatron.energon.SavableDatasetWrapper`."""

    #: The state of the torch random number generator
    torch_rng: bytes
    #: The state of the numpy random number generator
    numpy_rng: Tuple[str, bytes, int, int, float]
    #: The state of the python random number generator
    rng: Any
    #: Index of the next sample to be returned from the dataset
    sample_index: int

    def __repr__(self):
        tr = self.torch_rng[:3] + b"..."
        nr = (self.numpy_rng[0], self.numpy_rng[1][:3] + b"...", *self.numpy_rng[2:])
        r = (self.rng[0], self.rng[1] + ("...",), self.rng[2])
        return f"SavableDatasetState(torch_rng={tr!r}, numpy_rng={nr!r}, rng={r!r}, sample_index={self.sample_index})"


@dataclass
class SavableDatasetMergedState(MergedState):
    """Merged state of the dataset wrapper. See :class:`megatron.energon.SavableDatasetState` for more
    information. This class is not intended to be used directly, but by
    :class:`megatron.energon.SavableDatasetWrapper`."""

    #: The state of the torch random number generator
    torch_rng: List[bytes]
    #: The state of the numpy random number generator
    numpy_rng: List[Tuple[str, bytes, int, int, float]]
    #: The state of the python random number generator
    rng: List[Any]
    #: Index of the next sample to be returned from the dataset
    sample_index: List[int]

    def __repr__(self):
        tr = [None if r is None else r[:3] + b"..." for r in self.torch_rng]
        nr = [None if r is None else (r[0], r[1][:3] + b"...", *r[2:]) for r in self.numpy_rng]
        r = [None if r is None else (r[0], r[1][:3] + ("...",), r[2]) for r in self.rng]
        return f"SavableDatasetMergedState(torch_rng={tr!r}, numpy_rng={nr!r}, rng={r!r}, sample_index={self.sample_index})"


@dataclass
class SavableCheckpoint:
    """Checkpoint data for :class:`megatron.energon.SavableDatasetWrapper`. An instance is created
    regularly to be able to save the state of the dataset wrapper before the currently emitted
    sample.
    """

    #: The state of the wrapper
    state: Optional[SavableDatasetState]
    #: The state of the inner dataset
    dataset_state: Optional[State]
    #: The time at which the checkpoint was created
    checkpoint_time: float
    #: Index of the next sample to be returned from the dataset after restoring the checkpoint
    sample_index: int


@dataclass
class SavableDatasetCheckpoint(State):
    """Checkpoint data for :class:`megatron.energon.SavableDatasetWrapper`. The checkpoint state
    represents a state before that checkpoint, with an offset (i.e. samples to be skipped)."""

    #: The state of the wrapper at the sample index when the checkpoint was created.
    state: Optional[SavableDatasetState]
    #: The state of the inner dataset at the sample index when the checkpoint was created.
    dataset_state: Optional[State]
    #: Offset of the checkpoint to the actual sample index to be restored.
    offset: int


@dataclass
class SavableDatasetMergedCheckpoint(MergedState):
    """Checkpoint data for :class:`megatron.energon.SavableDatasetWrapper`. The checkpoint state
    represents a state before that checkpoint, with an offset (i.e. samples to be skipped)."""

    #: The state of the wrapper at the sample index when the checkpoint was created.
    state: Optional[SavableDatasetMergedState]
    #: The state of the inner dataset at the sample index when the checkpoint was created.
    dataset_state: Optional[MergedState]
    #: Offset of the checkpoint to the actual sample index to be restored.
    offset: List[int]


class SavableDatasetWrapper(IterableDataset[Tuple[int, int, T]], Generic[T]):
    """Internal class for wrapping a savable dataset for a worker process. Provides communication
    with the :class:`megatron.energon.SavableDataLoader`. This class is not intended to be used directly.
    See :class:`megatron.energon.SavableDataLoader` for more information."""

    #: The wrapped dataset
    dataset: SavableDataset[T]
    #: The configuration of the worker process
    worker_config: WorkerConfig
    #: The time interval in seconds to wait at minimum between two checkpoints
    checkpoint_every_sec: float
    #: The minimum number of samples to be emitted between two checkpoints. Should be `number of
    # workers * 2`.
    checkpoint_every_min_n_samples: int
    #: The number of checkpoints to keep in memory, before discarding. Should be 2.
    n_checkpoints: int
    #: The queue of the worker process to receive commands from the `SavableDataLoader`.
    _cmd_queues: List[torch.multiprocessing.Queue]
    #: The queue of the worker process to send results to the `SavableDataLoader`.
    _result_queues: List[torch.multiprocessing.Queue]

    _sample_index: int = 0
    _worker_offset: int = 0
    _last_checkpoints: List[SavableCheckpoint]
    _restore_from: Optional[Any] = None
    _skip_samples: List[int]

    _running: bool = False
    _command_lock: Optional[threading.RLock] = None
    _cmd_thread: Optional[threading.Thread] = None

    def __init__(
        self,
        dataset: SavableDataset[T],
        worker_config: WorkerConfig,
        checkpoint_every_sec: float,
        checkpoint_every_min_n_samples: int,
        n_checkpoints: int = 2,
        *,
        cmd_queues: List[torch.multiprocessing.Queue],
        result_queues: List[torch.multiprocessing.Queue],
    ):
        """
        Create the savable dataset wrapper for multiprocessing data loading.

        Args:
            dataset: The dataset to wrap
            worker_config: The worker config as used by all datasets
            checkpoint_every_sec: The time interval in seconds to wait at minimum between two
                checkpoints.
            checkpoint_every_min_n_samples: The minimum number of samples to be emitted between
                two checkpoints. Should be `number of workers * 2`.
            n_checkpoints: Number of checkpoints to keep.
            cmd_queues: The command queues for communicating with the worker processes.
            result_queues: The result queues for communicating with the worker processes.
        """
        self.dataset = dataset
        self.worker_config = worker_config
        self.checkpoint_every_sec = checkpoint_every_sec
        self.checkpoint_every_min_n_samples = checkpoint_every_min_n_samples
        self.n_checkpoints = n_checkpoints
        self._last_checkpoints = [
            SavableCheckpoint(
                state=None, dataset_state=None, checkpoint_time=time.perf_counter(), sample_index=0
            )
        ]
        self._skip_samples = [0] * max(worker_config.num_workers, 1)
        self._cmd_queues = cmd_queues
        self._result_queues = result_queues

    @staticmethod
    def _command_thread(self: "SavableDatasetWrapper"):
        """The internal thread, which processes the command and result queues. This thread is
        static, because `self` is actually passed as weakref proxy, to avoid keeping the dataset
        alive via the thread.
        """
        # print(f"{id(self)}:{multiprocessing.current_process().ident} Worker command thread starting")
        try:
            while self._running:
                try:
                    cmd_args = self._cmd_queues[self._worker_id].get(timeout=1)
                except queue.Empty:
                    continue
                # print(f"recv cmd {cmd_args}")
                with self._command_lock:
                    cmd = cmd_args[0]
                    if cmd is None:
                        break
                    try:
                        fn = getattr(self, cmd)
                        self._result_queues[self._worker_id].put(
                            {self._worker_id: fn(*cmd_args[1:])}
                        )
                        # print(f"result sent")
                    except Exception as e:
                        traceback.print_exc()
                        self._result_queues[self._worker_id].put({self._worker_id: e})
                        # print(f"exc sent")
        except BaseException:
            traceback.print_exc()
            raise
        finally:
            pass
            # print(f"{id(self)}:{multiprocessing.current_process().ident} Worker command thread closing")

    def __len__(self):
        return len(self.dataset)

    def __del__(self):
        if self._cmd_thread is not None:
            # print(f"{id(self)}:{multiprocessing.current_process().ident} Closing cmd thread")
            self._running = False
            self._cmd_thread.join()
            self._command_lock = None
            self._cmd_thread = None
            # print(f"{id(self)}:{multiprocessing.current_process().ident} Cmd thread closed")

    def __iter__(self):
        # First: Set the worker offset globally for the current worker
        WorkerConfig.worker_id_offset = self._worker_offset
        self._worker_id = self.worker_config.rank_worker_id()
        if self._cmd_thread is None:
            self._running = True
            self._command_lock = threading.RLock()
            weakref_self = weakref.proxy(self)
            self._cmd_thread = threading.Thread(
                target=SavableDatasetWrapper._command_thread,
                name="command_thread",
                args=(weakref_self,),
                daemon=True,
            )
            self._cmd_thread.start()
            # atexit.register(lambda: weakref_self.__del__())
        try:
            with self._command_lock:
                if self._restore_from:
                    self._restore_state(self._restore_from)
                    self._restore_from = None
                # If skipping, also restart the iterator to reach the start of the restored
                # checkpoint
                last_was_skip = True
                while last_was_skip:
                    dataset_has_samples = False
                    self.worker_config.worker_activate(self._sample_index)
                    try:
                        for src_data in self.dataset:
                            self.worker_config.worker_deactivate()
                            dataset_has_samples = True
                            if self._skip_samples[self._worker_id] > 0:
                                # Skip ahead to reach the start of the restored checkpoint
                                # print(f"Skip [{self._worker_id}:{self._sample_index}] {src_data}")
                                self._skip_samples[self._worker_id] -= 1
                                self._sample_index += 1
                                last_was_skip = True
                                self.worker_config.worker_activate(self._sample_index)
                                continue
                            last_was_skip = False
                            sample_index = self._sample_index
                            self._sample_index += 1
                            self._store_checkpoint()
                            try:
                                self._command_lock.release()
                                # print(f"{id(self)}:{multiprocessing.current_process().ident} Lock released")
                                # Commands may be executed only when data was yielded, not during
                                # iteration fetching.
                                # print(f"Yield next data [{self._worker_id}:{sample_index}] {src_data}")
                                yield self._worker_id, sample_index, src_data
                            finally:
                                # print(f"{id(self)}:{multiprocessing.current_process().ident} Lock acquiring")
                                self._command_lock.acquire()
                                # print(f"{id(self)}:{multiprocessing.current_process().ident} Lock acquired")
                            self.worker_config.worker_activate(self._sample_index)
                    finally:
                        self.worker_config.worker_deactivate()

                    # If the dataset is empty, don't try again and again
                    if not dataset_has_samples:
                        break
        finally:
            # print(f"{id(self)}:{multiprocessing.current_process().ident} Worker iter closing")
            # Always store a final checkpoint (it's likely to be saved)
            self._store_checkpoint(force=True)

    def _store_checkpoint(self, force: bool = False) -> None:
        """
        Internally create a checkpoint for the current state. This is required to store states
        from the past, which have already been yielded here, but not yet been retrieved from the
        intermediate queues.

        Args:
            force: If true, ignore time or frequency condition.
        """
        if (
            force
            or (
                self._last_checkpoints[-1].checkpoint_time + self.checkpoint_every_sec
                < time.perf_counter()
                and self._last_checkpoints[-1].sample_index + self.checkpoint_every_min_n_samples
                <= self._sample_index
            )
            or self._sample_index <= 1
        ):
            # print(f"Storing checkpoint at {self._worker_id}:{self._sample_index}")
            self._last_checkpoints.append(
                SavableCheckpoint(
                    state=self._save_state(),
                    dataset_state=self.dataset.save_state(),
                    checkpoint_time=time.perf_counter(),
                    sample_index=self._sample_index,
                )
            )
            if len(self._last_checkpoints) > self.n_checkpoints:
                self._last_checkpoints.pop(0)

    def _save_state(self) -> SavableDatasetState:
        """Saves the internal state"""
        (
            np_tp,
            np_state,
            pos,
            has_gauss,
            cached_gaussian,
        ) = np.random.get_state()
        return SavableDatasetState(
            torch_rng=bytes(torch.get_rng_state().tolist()),
            numpy_rng=(np_tp, np_state.tobytes(), pos, has_gauss, cached_gaussian),
            rng=random.getstate(),
            sample_index=self._sample_index,
        )

    def _merge_states(self, states: List[SavableDatasetState]) -> SavableDatasetMergedState:
        """Merges the internal state"""
        assert torch.utils.data.get_worker_info() is None, "Cannot merge in worker process"
        return SavableDatasetMergedState(
            torch_rng=[None if s is None else s.torch_rng for s in states],
            numpy_rng=[None if s is None else s.numpy_rng for s in states],
            rng=[None if s is None else s.rng for s in states],
            sample_index=[0 if s is None else s.sample_index for s in states],
        )

    def _restore_state(self, state: SavableDatasetMergedState) -> None:
        """Restores the internal worker state"""
        assert torch.utils.data.get_worker_info() is not None, "Can only restore in worker process"
        if state.torch_rng[self._worker_id] is None:
            torch.manual_seed(torch.initial_seed())
        else:
            torch.set_rng_state(
                torch.frombuffer(
                    bytearray(state.torch_rng[self._worker_id]), dtype=torch.uint8
                ).clone()
            )
        if state.numpy_rng[self._worker_id] is None:
            np.random.seed(torch.initial_seed() & 0xFFFFFFFF)
        else:
            np_tp, np_state, *np_rng = state.numpy_rng[self._worker_id]
            np.random.set_state((np_tp, np.frombuffer(np_state, dtype=np.uint32), *np_rng))
        if state.rng[self._worker_id] is None:
            random.seed(torch.initial_seed() & 0xFFFFFFFF)
        else:
            random.setstate(state.rng[self._worker_id])
        self._sample_index = state.sample_index[self._worker_id]
        self._last_checkpoints = [
            SavableCheckpoint(
                state=self._save_state(),
                dataset_state=self.dataset.save_state(),
                checkpoint_time=time.perf_counter(),
                sample_index=self._sample_index,
            )
        ]

    def get_checkpoint(self, last_sample_indexes: List[int]) -> SavableDatasetCheckpoint:
        """
        Get a checkpoint given the last emitted sample indexes for all workers.

        Args:
            last_sample_indexes: The last emitted sample indexes for all workers.

        Returns:
            The found checkpoint including the offset to the next sample index
        """
        sample_index = last_sample_indexes[self._worker_id] + 1
        for checkpoint in reversed(self._last_checkpoints):
            if checkpoint.sample_index <= sample_index:
                # print(f"Found cp for {sample_index} at {checkpoint.sample_index}")
                return SavableDatasetCheckpoint(
                    dataset_state=checkpoint.dataset_state,
                    state=checkpoint.state,
                    offset=sample_index - checkpoint.sample_index,
                )
        raise ValueError("No checkpoint found")

    def merge_checkpoints(
        self, checkpoints: List[SavableDatasetCheckpoint]
    ) -> SavableDatasetMergedCheckpoint:
        """Merges saved checkpoints from all worker processes."""
        assert torch.utils.data.get_worker_info() is None, "Cannot merge in worker process"
        assert all(isinstance(c, SavableDatasetCheckpoint) for c in checkpoints)
        return SavableDatasetMergedCheckpoint(
            dataset_state=self.dataset.merge_states([c.dataset_state for c in checkpoints]),
            state=self._merge_states([c.state for c in checkpoints]),
            offset=[c.offset for c in checkpoints],
        )

    def restore_checkpoint(
        self,
        state: Optional[SavableDatasetMergedCheckpoint],
        worker_offset: int,
    ) -> None:
        """
        Restores the merged checkpoint from all worker processes.

        Args:
            state: The state to restore
            worker_offset: The offset of the last worker which has emitted a sample. This will be
                set in all worker processes to ensure the right worker starts as first.
        """
        assert torch.utils.data.get_worker_info() is None, "Cannot restore in worker process"
        if state is None:
            self._restore_from = None
            assert worker_offset == 0
            self._worker_offset = 0
            self.dataset.restore_state(None)
            self._skip_samples = [0] * max(self.worker_config.num_workers, 1)
        else:
            assert isinstance(state, SavableDatasetMergedCheckpoint)
            self._restore_from = state.state
            self._worker_offset = worker_offset
            self.dataset.restore_state(state.dataset_state)
            self._skip_samples = state.offset

    def config(self) -> Dict[str, Any]:
        return self.dataset.config()

    def __str__(self):
        return f"SavableDatasetWrapper(dataset={self.dataset})"


@dataclass
class SavableDataLoaderState(State):
    """Saved state of the :class:`megatron.energon.SavableDataLoader`. Contains the state for all worker
    processed of a single rank."""

    #: The internal state of the dataset (i.e. for worker processes, this will be a
    # :class:`megatron.energon.SavableDatasetMergedState`, otherwise directly the inner dataset state)
    dataset_state: Union[SavableDatasetMergedCheckpoint, MergedState]
    #: Which worker will be the next to emit a sample. Used to restore the proper order
    next_worker_id: int


class SavableDataLoader(DataLoader[T], Generic[T]):
    """DataLoader that supports saving and restoring the state of the dataset.
    When restoring, the dataloader and dataset must be instantiated with the exactly same
    parameters.

    ## How this works (for no worker processes)
    1. The state of the dataset is saved using :meth:`megatron.energon.SavableDataset.save_state`
    2. (for compatibility) The state of the dataset is converted to using inner arrays using
       :meth:`megatron.energon.SavableDataset.merge_states`.
    3. The state can be restored using :meth:`megatron.energon.SavableDataset.restore_state` given the
       previously saved (and merged) state.

    ## How this works (for worker processes)
    - First issue is, that worker processes work with internal queues between processes to pass
      loaded samples to the main process (also to perform collating). This means that the whole
      state of the dataset is not directly accessible from the main process.
    - To solve this issue, the dataset regularly saves a checkpoint of its state to be able to
      resume from that state (and skip the samples that have already been yielded).
    - To have a consistent state, the sample index from the latest yielded samples is saved for all
      worker instances. Thus, the main process knows exactly which sample indexes should come next
      from which worker.
    - Internally, pytorch iterates through the workers in order to retrieve the next worker's
      samples. Unfortunately, that next worker index cannot be restored in pytorch's dataloader,
      thus the workers are shifted internally by that offset
      (see :attr:`megatron.energon.WorkerConfig.worker_id_offset`).

    1. The dataset is wrapped in a :class:`megatron.energon.SavableDatasetWrapper`. This allows the main
       process to communicate with the worker and send commands to the workers and retrieve the
       results.
    2. The state of the dataset is saved using
       :meth:`megatron.energon.SavableDatasetWrapper.get_checkpoint`. This gives the last checkpoint
       from the requested sample index and stores the offset (i.e. number of samples to skip) from
       that checkpoint.
    3. The state is merged using :meth:`megatron.energon.SavableDatasetWrapper.merge_checkpoints`. This
       merges the states of all workers and returns a single state that can be used to restore the
       state of the dataset.
    3. The state can be restored using :meth:`megatron.energon.SavableDatasetWrapper.restore_state`
       before a worker is started, such that all workers initially receive the same state array.
       The worker firstly sets the worker index offset, then uses its (shifted) own index to get its
       required state from the merged state array.
    """

    #: The worker config
    worker_config: WorkerConfig
    #: The wrapped dataset. For multiprocessing, this is a :class:`megatron.energon.SavableDatasetWrapper`
    dataset: Union[SavableDatasetWrapper[T], SimpleSavableDatasetWrapper[T]]

    #: The global ID counter
    _next_id: ClassVar[int] = 0
    #: Class instance id
    id: int = 0

    #: The queues used to send commands to the workers
    cmd_queues: List[torch.multiprocessing.Queue]
    #: The queues used to receive results from the workers
    result_queues: List[torch.multiprocessing.Queue]

    #: Instance of the current data iterator. There shall be only one active iterator, such that the
    # dataset is not iterated multiple times in parallel. The state will proceed.
    _persistent_iterator: Optional[Iterator[T]] = None
    #: The index of the current worker
    _worker_sample_counters: List[int]
    #: Id of the next worker to retrieve data from
    _next_worker_id: int = 0
    #: Global index of the last yielded sample
    _global_sample_idx: int = 0
    #: Current iterator index of the last yielded sample
    _sample_idx: int = 0

    def __init__(
        self,
        dataset: SavableDataset[T],
        *,
        worker_config: Optional[WorkerConfig] = None,
        checkpoint_every_sec: float = 60,
        checkpoint_every_min_n_samples: Optional[int] = None,
        n_checkpoints: int = 2,
    ):
        """
        Create the dataloader supporting saving and restoring the state.

        Args:
            dataset: The dataset to load.
            worker_config: The worker config to use
            checkpoint_every_sec: This is the time in seconds after which a checkpoint is saved.
                It may take the same duration to restore a checkpoint, but introduces additional
                overhead during reading data from the dataset, so this should be chosen accordingly.
                Only applies if using workers.
            checkpoint_every_min_n_samples: Overwrites the minimum number of samples between
                checkpoints. Defaults to `number of workers * 2`. Only applies if using workers.
            n_checkpoints: The number of checkpoints to keep in memory. Only applies if using
                workers.
        """
        self.worker_config = worker_config or global_worker_config
        self.id = self.next_id()

        dataset = GcDataset(dataset)

        self.cmd_queues = [multiprocessing.Queue() for _ in range(self.worker_config.num_workers)]
        self.result_queues = [
            multiprocessing.Queue() for _ in range(self.worker_config.num_workers)
        ]

        if self.worker_config.num_workers > 0:
            if checkpoint_every_min_n_samples is None:
                checkpoint_every_min_n_samples = self.worker_config.num_workers * 2
            dataset = SavableDatasetWrapper(
                dataset,
                self.worker_config,
                checkpoint_every_sec=checkpoint_every_sec,
                checkpoint_every_min_n_samples=checkpoint_every_min_n_samples,
                n_checkpoints=n_checkpoints,
                cmd_queues=self.cmd_queues,
                result_queues=self.result_queues,
            )
        else:
            dataset = SimpleSavableDatasetWrapper(dataset, self.worker_config)

        self._worker_sample_counters = [0] * max(self.worker_config.num_workers, 1)

        kwargs = {}
        if self.worker_config.num_workers > 0:
            kwargs["persistent_workers"] = True
            kwargs["prefetch_factor"] = 2

        # Compute seeds for each worker, based on current rank
        seed_per_worker = [
            self.worker_config.worker_seed(i) for i in range(self.worker_config.num_workers)
        ]

        super().__init__(
            dataset,
            batch_size=None,
            shuffle=False,
            num_workers=self.worker_config.num_workers,
            pin_memory=True,
            worker_init_fn=partial(_init_worker, seed_per_worker),
            **kwargs,
        )

        if self.worker_config.should_log(level=1):
            self.worker_config.worker_log(
                {
                    "t": "SavableLoader.__init__",
                    "r": self.worker_config.rank,
                    "w": None,
                    "id": self.id,
                    "config": dataset.config(),
                }
            )

    @staticmethod
    def next_id() -> int:
        next_id = SavableDataLoader._next_id
        SavableDataLoader._next_id += 1
        return next_id

    def __iter__(self):
        outerself = self

        class InnerIterator:
            """Internal class which keeps the iterator alive across multiple `iter()` calls.
            If the inner iterator is exhausted, will also exhaust and a new instance is needed.
            Also saves the last sample index and the next worker id.
            """

            finished: bool = False
            iter_idx: int = 0
            id: int

            def __init__(self, iterator):
                self._iterator = iterator
                self.id = outerself.next_id()
                if outerself.worker_config.should_log(level=1):
                    outerself.worker_config.worker_log(
                        {
                            "t": "SavableDataLoader.iter",
                            "r": outerself.worker_config.rank,
                            "w": None,
                            "id": outerself.id,
                            "iter_id": self.id,
                        }
                    )

                # self._debugf = open(
                #     f"worker_samples_rank{outerself.worker_config.rank:02}_t{int(time.time())}.log", "w"
                # )

            def __iter__(self):
                return self

            def __next__(self):
                try:
                    worker_id, sample_idx, sample = next(self._iterator)
                    outerself._worker_sample_counters[worker_id] = sample_idx
                    # If the next sample will be from the first worker, we can safely resume
                    outerself._next_worker_id = (worker_id + 1) % max(outerself.num_workers, 1)
                    # self._debugf.write(
                    #     f"[w={worker_id}, s={sample_idx}] {self._sample_str(sample)}\n"
                    # )
                    # self._debugf.flush()
                    if outerself.worker_config.should_log(level=1):
                        keys = default_get_keys(sample)
                        outerself.worker_config.worker_log(
                            {
                                **{
                                    "t": "SavableDataLoader.yield",
                                    "r": outerself.worker_config.rank,
                                    "w": None,
                                    "id": outerself.id,
                                    "iter_id": self.id,
                                    "worker_id": worker_id,
                                    "worker_idx": sample_idx,
                                    "idx": outerself._sample_idx,
                                    "iter_idx": self.iter_idx,
                                    "global_idx": outerself._global_sample_idx,
                                },
                                **({} if keys is None else {"keys": keys}),
                            }
                        )
                    outerself._sample_idx += 1
                    outerself._global_sample_idx += 1
                    self.iter_idx += 1
                    return sample
                except StopIteration:
                    self.finished = True
                    outerself._next_worker_id = 0
                    if outerself.worker_config.should_log(level=1):
                        outerself.worker_config.worker_log(
                            {
                                "t": "SavableDataLoader.StopIteration",
                                "r": outerself.worker_config.rank,
                                "w": None,
                                "id": outerself.id,
                                "iter_id": self.id,
                            }
                        )
                    raise

        if self.num_workers > 0:
            # Always keep same iterator alive, as long as it yields data
            if self._persistent_iterator is None or self._persistent_iterator.finished:
                self._persistent_iterator = InnerIterator(super().__iter__())
                self._sample_idx = 0
                # print("New Iterator", self._persistent_iterator)
            return self._persistent_iterator
        else:
            return InnerIterator(super().__iter__())

    def _worker_command(self, *cmd_args) -> List[Any]:
        """Executes a command in all workers and returns the results."""
        # print(f"cmd: {cmd_args}")
        for cmd_queue in self.cmd_queues:
            cmd_queue.put(cmd_args)
        # print(f"waiting for res")
        assert len(self.result_queues) == self.worker_config.num_workers
        res = {k: v for results_queue in self.result_queues for k, v in results_queue.get().items()}
        res = [res[i] for i in range(len(res))]
        # print(f"res: {res}")
        for r in res:
            if isinstance(r, Exception):
                raise r
        return res

    def save_state_rank(self) -> Optional[SavableDataLoaderState]:
        """
        Saves the state of the dataset for the current rank. Allows for restoring the state later
        using `restore_state_rank`, given the result of this method.

        Returns:
            The state of the dataset.
        """
        # Fetch current rank's worker's state
        if self.num_workers == 0:
            # No workers configured
            worker_states = [self.dataset.save_state()]
            assert self._next_worker_id == 0
            merged_states = self.dataset.merge_states(worker_states)
        elif self._persistent_iterator is None:
            # Workers configured, but not started yet -> Initial state
            return None
        else:
            # Fetch from worker processes
            worker_states = self._worker_command("get_checkpoint", self._worker_sample_counters)
            merged_states = self.dataset.merge_checkpoints(worker_states)

        # Merge the states
        merged_state = SavableDataLoaderState(
            dataset_state=merged_states,
            next_worker_id=self._next_worker_id,
        )
        # print("Merged state", merged_state)

        # Not distributed -> return the merged state
        return merged_state

    def restore_state_rank(self, state: Optional[SavableDataLoaderState]) -> None:
        """
        Restores the saved state for the current rank.

        Args:
            state: The state to restore, as saved by `save_state_rank`.
        """
        assert self._persistent_iterator is None, "Cannot restore state while workers are running"
        if state is None:
            # Assume initial state
            return
        assert isinstance(state, SavableDataLoaderState)
        if isinstance(self.dataset, SavableDataset):
            self.dataset.restore_state(state.dataset_state)
        else:
            assert isinstance(self.dataset, SavableDatasetWrapper)
            assert isinstance(state.dataset_state, SavableDatasetMergedCheckpoint)
            self.dataset.restore_checkpoint(state.dataset_state, worker_offset=state.next_worker_id)
        # if len(self.cmd_queues) > 0:
        #     self._worker_command("restore_state", state["dataset"])

    def save_state(
        self, dst_rank: Optional[int] = None
    ) -> Union[SavableDataLoaderState, List[SavableDataLoaderState], None, List[None]]:
        """
        Saves the state of the dataset. Allows for restoring the state later using `restore_state`,
        given the result of this method.

        Args:
            dst_rank: If specified, the state will be gathered to this rank. Otherwise, it will be
                gathered to all ranks.

        Returns:
            The state of the dataset (or `None`, if not on `dst_rank`).
        """
        # Fetch current rank's worker's state
        merged_state = self.save_state_rank()
        # print("Merged state", merged_state)

        # Gather the merged states
        if torch.distributed.is_available() and torch.distributed.is_initialized():
            output: Optional[List[Optional[MergedState]]]
            if dst_rank is None:
                output = [None] * self.worker_config.world_size
                torch.distributed.all_gather_object(
                    output, merged_state, group=self.worker_config.data_parallel_group
                )
            else:
                if self.worker_config.rank == dst_rank:
                    output = [None] * self.worker_config.world_size
                else:
                    output = None
                torch.distributed.gather_object(
                    merged_state, output, dst_rank, group=self.worker_config.data_parallel_group
                )
            return output
        else:
            # Not distributed -> return the merged state
            return merged_state

    def restore_state(
        self,
        state: Union[SavableDataLoaderState, List[SavableDataLoaderState], None, List[None]],
        src_rank: Optional[int] = None,
    ) -> None:
        """
        Restores the saved state.

        Args:
            state: The state to restore, as saved by `save_state`.
            src_rank: If set, only this rank is assumed to hold the data, other ranks must set state
                to None. If not set, all ranks are assumed to hold the data.
        """
        assert self._persistent_iterator is None, "Cannot restore state while workers are running"
        # Only restore multi-rank if state is actually a list and we are in a torch distributed setup.
        # Otherwise treat as single rank state.
        if (
            torch.distributed.is_available()
            and torch.distributed.is_initialized()
            and isinstance(state, list)
        ):
            if src_rank is None:
                # All ranks have the state
                # Select the state of the current rank
                state = state[self.worker_config.rank]
            else:
                # Only the src_rank has the state
                if self.worker_config.rank != src_rank:
                    # Send the state to all other ranks
                    assert state is None
                    # Must still be a list of Nones
                    state = [None] * self.worker_config.world_size
                local_object = [None]
                torch.distributed.scatter_object_list(
                    local_object,
                    state,
                    src=src_rank,
                    group=self.worker_config.data_parallel_group,
                )
                state = local_object[0]
        self.restore_state_rank(state)

    def config(self):
        """Get the configuration, which defines the dataset. Useful in conjunction with `save_state`
        and `restore_state` to match the configuration as well."""
        return {
            "type": type(self).__qualname__,
            "num_workers": self.num_workers,
            "persistent_workers": self.persistent_workers,
            "pin_memory": self.pin_memory,
            "prefetch_factor": None if self.num_workers == 0 else self.prefetch_factor,
            "dataset": self.dataset.config(),
        }


class BasicDataLoader(DataLoader[T], Generic[T]):
    """DataLoader that supports debugging the dataset without saving capability (e.g. for val/eval)."""

    #: The worker config
    worker_config: WorkerConfig
    #: The wrapped dataset. For multiprocessing, this is a :class:`megatron.energon.SavableDatasetWrapper`
    dataset: Union[SavableDatasetWrapper[T], SavableDataset[T]]

    id: int
    _sample_idx: int = 0

    def __init__(
        self,
        dataset: SavableDataset[T],
        *,
        worker_config: Optional[WorkerConfig] = None,
    ):
        """
        Create the dataloader supporting saving and restoring the state.

        Args:
            dataset: The dataset to load.
            worker_config: The worker config to use
            checkpoint_every_sec: This is the time in seconds after which a checkpoint is saved.
               It may take the same duration to restore a checkpoint, but introduces additional
               overhead during reading data from the dataset, so this should be chosen accordingly.
        """
        self.worker_config = worker_config or global_worker_config

        self.id = SavableDataLoader.next_id()

        dataset = GcDataset(dataset)
        dataset = SimpleSavableDatasetWrapper(dataset, self.worker_config)

        self._worker_sample_counters = [0] * max(self.worker_config.num_workers, 1)

        kwargs = {}
        if self.worker_config.num_workers > 0:
            # These must not be specified for num_workers =0
            kwargs["persistent_workers"] = True
            kwargs["prefetch_factor"] = 2

        seed_per_worker = [
            self.worker_config.worker_seed(i) for i in range(self.worker_config.num_workers)
        ]

        gc.collect()  # This ensures that we don't include any old worker refs in the newly forked worker processes

        super().__init__(
            dataset,
            batch_size=None,
            shuffle=False,
            num_workers=self.worker_config.num_workers,
            pin_memory=True,
            worker_init_fn=partial(_init_worker, seed_per_worker),
            **kwargs,
        )
        if self.worker_config.should_log(level=1):
            self.worker_config.worker_log(
                {
                    "t": "BasicDataLoader.__init__",
                    "r": self.worker_config.rank,
                    "w": None,
                    "id": self.id,
                    "config": self.config(),
                }
            )

    def __iter__(self):
        outerself = self

        class InnerIterator:
            """Internal class which keeps the iterator alive across multiple `iter()` calls.
            If the inner iterator is exhausted, will also exhaust and a new instance is needed.
            Also saves the last sample index and the next worker id.
            """

            iter_idx: int = 0
            id: int

            def __init__(self, iterator):
                self._iterator = iterator
                self.id = SavableDataLoader.next_id()

                if outerself.worker_config.should_log(level=1):
                    outerself.worker_config.worker_log(
                        {
                            "t": "BasicDataLoader.iter",
                            "r": outerself.worker_config.rank,
                            "w": None,
                            "id": outerself.id,
                            "iter_id": self.id,
                        }
                    )

            def __iter__(self):
                return self

            def __next__(self):
                try:
                    worker_id, sample_idx, sample = next(self._iterator)
                    # If the next sample will be from the first worker, we can safely resume
                    self.next_worker_id = (worker_id + 1) % max(outerself.num_workers, 1)
                    if outerself.worker_config.should_log(level=1):
                        keys = default_get_keys(sample)
                        outerself.worker_config.worker_log(
                            {
                                **{
                                    "t": "BasicDataLoader.yield",
                                    "r": outerself.worker_config.rank,
                                    "w": None,
                                    "id": outerself.id,
                                    "iter_id": self.id,
                                    "worker_id": worker_id,
                                    "worker_idx": sample_idx,
                                    "idx": self.iter_idx,
                                    "iter_idx": self.iter_idx,
                                    "global_idx": outerself._sample_idx,
                                },
                                **({} if keys is None else {"keys": keys}),
                            }
                        )
                    outerself._sample_idx += 1
                    self.iter_idx += 1
                    return sample
                except StopIteration:
                    self.next_worker_id = 0
                    if outerself.worker_config.should_log(level=1):
                        outerself.worker_config.worker_log(
                            {
                                "t": "BasicDataLoader.StopIteration",
                                "r": outerself.worker_config.rank,
                                "w": None,
                                "id": outerself.id,
                                "iter_id": self.id,
                            }
                        )
                    raise

        return InnerIterator(super().__iter__())

    def config(self):
        """Get the configuration, which defines the dataset. Useful in conjunction with `save_state`
        and `restore_state` to match the configuration as well."""
        return {
            "type": type(self).__qualname__,
            "num_workers": self.num_workers,
            "persistent_workers": self.persistent_workers,
            "pin_memory": self.pin_memory,
            "prefetch_factor": None if self.num_workers == 0 else self.prefetch_factor,
            "dataset": self.dataset.config(),
        }


def _sample_str(self, sample):
    """Returns a human readable debug string for a single sample, also uniquely identifying it."""
    import dataclasses
    import hashlib

    if isinstance(sample, torch.Tensor):
        return f"Tensor(shape={sample.shape}, dtype={sample.dtype}, sha256={hashlib.sha256(sample.detach().cpu().numpy().tobytes()).hexdigest()!r})"
    elif isinstance(sample, np.ndarray):
        return f"ndarray(shape={sample.shape}, dtype={sample.dtype}, sha256={hashlib.sha256(sample.tobytes()).hexdigest()!r})"
    elif isinstance(sample, (str, bytes)):
        if len(sample) > 100:
            return f"{sample[:100]!r}..."
        return repr(sample)
    elif isinstance(sample, (int, float)):
        return repr(sample)
    elif isinstance(sample, dict):
        innerstr = ", ".join(f"{k}={self._sample_str(v)}, " for k, v in sample.items())
        return f"dict({innerstr})"
    elif isinstance(sample, (list, tuple)):
        innerstr = ", ".join(self._sample_str(v) for v in sample)
        return f"{type(sample)}({innerstr})"
    elif dataclasses.is_dataclass(sample):
        innerstr = ", ".join(
            f"{field.name}={self._sample_str(getattr(sample, field.name))}, "
            for field in dataclasses.fields(sample)
        )
        return f"{type(sample).__name__}({innerstr})"
    else:
        return repr(sample)



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause
"""Provides methods for converting typed objects to json objects and vice versa."""

import dataclasses
import importlib
import inspect
import json
import re
import typing
from typing import Any, Callable, Dict, Literal, Optional, Tuple, Type, TypeVar, Union

from megatron.energon.epathlib import EPath


class JsonValueError(ValueError):
    """Internal error class, for better error messages."""

    def __init__(
        self, msg: str, expected_type: Type, value: Any, path: str, stage: Tuple[int, ...]
    ) -> None:
        """New error"""
        super().__init__(msg)
        self.expected_type = expected_type
        self.value = value
        self.path = path
        self.stage = stage


class _MissingValue:
    """Internal class representing a missing value (i.e. a field which has not been specified,
    opposed to explicit `None`) for better error messages."""

    def __repr__(self) -> str:
        return "missing value"


_missing_value = _MissingValue()

TType = TypeVar("TType")


def _check_instance_type(cls, inst_type: Type) -> bool:
    """Check if a class is an instance of a type."""
    if inst_type is None:
        return True
    elif typing.get_origin(inst_type) is not None:
        org = typing.get_origin(inst_type)
        if org is Union:
            for check_type in typing.get_args(inst_type):
                if _check_instance_type(cls, check_type):
                    return True
        elif isinstance(org, type) and issubclass(cls, org):
            return True
    elif inst_type is Any:
        return True
    else:
        return not issubclass(inst_type, type) or issubclass(cls, inst_type)


def raw_to_instance(kwargs: dict, inst_type: Type[TType], strict: bool = True) -> TType:
    """
    Try to import and instantiate a class from a dict with "__module__" and "__class__" keys.

    Args:
        kwargs: The dict to parse
        inst_type: Expected return type, used if type is not specified in the kwargs
        strict: If true, don't allow additional attributes

    Returns:
        Instantiated class
    """
    kwargs = kwargs.copy()
    module_name = kwargs.pop("__module__", None)
    class_name = kwargs.pop("__class__", None)
    if module_name is None or class_name is None:
        cls = inst_type
    else:
        try:
            module = importlib.import_module(module_name)
        except ModuleNotFoundError:
            try_energon_module = importlib.import_module("megatron.energon", package=None)
            if hasattr(try_energon_module, class_name):
                module = try_energon_module
            else:
                raise

        cls = typing.cast(Type[TType], getattr(module, class_name))
        if isinstance(cls, type):
            assert _check_instance_type(cls, inst_type), f"Expected {inst_type}, got {cls}"
        elif not callable(cls):
            raise ValueError(f"Expected a class or a callable, got {cls}")
    inst = safe_call_function(kwargs, cls, strict=strict, allow_imports=True)
    if not isinstance(cls, type):
        assert _check_instance_type(
            type(inst), inst_type
        ), f"Expected {inst_type}, got {type(inst)}"
    return inst


def raw_to_typed(  # noqa: C901
    raw_data: Union[dict, list, str, int, bool, float, None],
    inst_type: Type[TType],
    strict: bool = False,
    allow_imports: bool = False,
    _path: str = "root",
    _stage: Tuple[int, ...] = (),
) -> TType:
    """
    Converts raw data (i.e. dicts, lists and primitives) to typed objects (like
    `NamedTuple` or `dataclasses.dataclass`). Validates that python typing matches.

    Usage::

        class MyNamedTuple(NamedTuple):
            x: int
            y: str

        assert raw_to_typed({'x': int, 'y': "foo"}, MyNamedTuple) == MyNamedTuple(x=5, y="foo")

    Args:
        raw_data: The raw (e.g. json) data to be made as `inst_type`
        inst_type: The type to return
        strict: If true, don't allow additional attributes
        allow_imports: If true, parse '__module__' and '__class__' attributes to allow explicit
            instantiation of types
        _path: (internal for recursive call) The path to the object being converted from the root
        _stage: (internal for recursive call) Numbers representing the position of the current
            object being converted from the root

    Returns:
        The input data as `inst_type`.
    """
    type_name = getattr(inst_type, "__name__", repr(inst_type))
    if raw_data is _missing_value:
        raise JsonValueError(
            f"Missing value at {_path}",
            inst_type,
            raw_data,
            _path,
            _stage,
        )
    elif inst_type in (str, int, float, bool, None, type(None)):
        # Literal types or missing data
        if not isinstance(raw_data, inst_type) and not (
            isinstance(raw_data, int) and inst_type == float
        ):
            raise JsonValueError(
                f"Type does not match, expected {type_name} at {_path}, got {raw_data!r}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
        return raw_data
    elif inst_type is Any:
        if (
            allow_imports
            and isinstance(raw_data, dict)
            and "__module__" in raw_data
            and "__class__" in raw_data
        ):
            return raw_to_instance(raw_data, inst_type, strict=strict)
        # Any
        return raw_data
    elif typing.get_origin(inst_type) is Literal:
        # Literal[value[, ...]]
        values = typing.get_args(inst_type)
        if raw_data not in values:
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {raw_data!r}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
        return raw_data
    elif typing.get_origin(inst_type) is Union:
        # Union[union_types[0], union_types[1], ...]
        union_types = typing.get_args(inst_type)
        if None in union_types:
            # Fast Optional path
            if raw_data is None:
                return None
        best_inner_error: Optional[JsonValueError] = None
        multiple_matching = False
        for subtype in union_types:
            try:
                return raw_to_typed(
                    raw_data,
                    subtype,
                    strict,
                    allow_imports,
                    f"{_path} -> {getattr(subtype, '__name__', repr(subtype))}",
                    _stage + (1,),
                )
            except JsonValueError as err:
                if best_inner_error is None or err.stage > best_inner_error.stage:
                    best_inner_error = err
                    multiple_matching = False
                elif err.stage == best_inner_error.stage:
                    multiple_matching = True
                continue
        if best_inner_error is None or multiple_matching:
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {raw_data!r}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
        else:
            raise best_inner_error
    elif (
        isinstance(inst_type, type)
        and issubclass(inst_type, tuple)
        and hasattr(inst_type, "__annotations__")
    ):
        # class MyClass(NamedTuple): ...
        if not isinstance(raw_data, dict):
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {raw_data!r}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
        if getattr(inst_type, "__dash_keys__", "False"):
            raw_data = {key.replace("-", "_"): val for key, val in raw_data.items()}
        defaults = getattr(inst_type, "_field_defaults", {})
        kwargs = {
            field_name: raw_to_typed(
                raw_data.get(field_name, defaults.get(field_name, _missing_value)),
                field_type,
                strict,
                allow_imports,
                f"{_path} -> {type_name}:{field_name}",
                _stage + (idx,),
            )
            for idx, (field_name, field_type) in enumerate(inst_type.__annotations__.items())
        }
        if strict and not set(raw_data).issubset(inst_type.__annotations__):
            raise JsonValueError(
                f"Additional attributes for {type_name} at {_path}, got {raw_data!r}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
        try:
            return inst_type(**kwargs)
        except BaseException:
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {raw_data!r}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
    elif dataclasses.is_dataclass(inst_type):
        # dataclass
        if not isinstance(raw_data, dict):
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {raw_data!r}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
        kwargs = {
            field.name: raw_to_typed(
                raw_data.get(
                    field.name,
                    (
                        (
                            _missing_value
                            if field.default_factory is dataclasses.MISSING
                            else field.default_factory()
                        )
                        if field.default is dataclasses.MISSING
                        else field.default
                    ),
                ),
                field.type,
                strict,
                allow_imports,
                f"{_path} -> {type_name}:{field.name}",
                _stage + (idx,),
            )
            for idx, field in enumerate(dataclasses.fields(inst_type))
            if field.init
        }
        if strict and not set(raw_data).issubset(
            field.name for field in dataclasses.fields(inst_type) if field.init
        ):
            raise JsonValueError(
                f"Additional attributes for {type_name} at {_path}, got {raw_data!r}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
        try:
            return inst_type(**kwargs)
        except BaseException:
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {raw_data!r}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
    elif typing.get_origin(inst_type) is list:
        # List[inner_type]
        (inner_type,) = typing.get_args(inst_type)
        if not isinstance(raw_data, list):
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {raw_data!r}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
        return [
            raw_to_typed(
                val, inner_type, strict, allow_imports, f"{_path} -> {idx}", _stage + (idx,)
            )
            for idx, val in enumerate(raw_data)
        ]
    elif typing.get_origin(inst_type) is set:
        # Set[inner_type]
        (inner_type,) = typing.get_args(inst_type)
        if not isinstance(raw_data, list):
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {raw_data!r}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
        res = set(
            raw_to_typed(
                val, inner_type, strict, allow_imports, f"{_path} -> {idx}", _stage + (idx,)
            )
            for idx, val in enumerate(raw_data)
        )
        if len(res) != len(raw_data):
            raise JsonValueError(
                f"Duplicate element at {_path}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
        return res
    elif typing.get_origin(inst_type) is tuple:
        # Tuple[inner_types[0], inner_types[1], ...] or Tuple[inner_types[0], Ellipsis/...]
        inner_types = typing.get_args(inst_type)
        if not isinstance(raw_data, list):
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {raw_data!r}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
        if len(inner_types) == 2 and inner_types[1] is Ellipsis:
            # Tuple of arbitrary length, all elements same type
            # Tuple[inner_types[0], Ellipsis/...]
            return tuple(
                raw_to_typed(
                    val, inner_types[0], strict, allow_imports, f"{_path} -> {idx}", _stage + (idx,)
                )
                for idx, val in enumerate(raw_data)
            )
        else:
            # Fixed size/typed tuple
            # Tuple[inner_types[0], inner_types[1], ...]
            if len(raw_data) != len(inner_types):
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
            return [
                raw_to_typed(
                    val, inner_type, strict, allow_imports, f"{_path} -> {idx}", _stage + (idx,)
                )
                for idx, (val, inner_type) in enumerate(zip(raw_data, inner_types))
            ]
    elif typing.get_origin(inst_type) is dict:
        # Dict[str, value_type]
        key_type, value_type = typing.get_args(inst_type)
        assert key_type is str
        if not isinstance(raw_data, dict):
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {raw_data!r}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
        return {
            key: raw_to_typed(
                val, value_type, strict, allow_imports, f"{_path} -> {key!r}", _stage + (idx,)
            )
            for idx, (key, val) in enumerate(raw_data.items())
        }
    elif inst_type in (dict, list):
        # dict, list (no subtyping)
        if not isinstance(raw_data, inst_type):
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {raw_data!r}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
        return raw_data
    elif inst_type == EPath:
        if isinstance(raw_data, str):
            return EPath(raw_data)
        elif not isinstance(raw_data, EPath):
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {raw_data!r}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
        return raw_data
    elif (
        allow_imports
        and isinstance(raw_data, dict)
        and "__module__" in raw_data
        and "__class__" in raw_data
    ):
        return raw_to_instance(raw_data, inst_type, strict=strict)
    else:
        return raw_data


def safe_call_function(
    raw_data: Union[dict, list, str, int, bool, float, None],
    fn: Callable[..., TType],
    strict: bool = False,
    allow_imports: bool = False,
) -> TType:
    """
    Converts raw data (i.e. dicts, lists and primitives) to typed call arguments.
    Validates that python typing matches.

    Usage::

        def fn(arg1: float, arg2: MyType, arg3) -> Any:
            assert isinstance(arg1, float)
            assert isinstance(arg2, MyType)

        fn(3.141, MyType(), None)

    Args:
        raw_data: The raw (e.g. json) data to be made as `inst_type`
        fn: The function to call with the converted data
        strict: If true, don't allow additional attributes
        allow_imports: If true, allow instantiating objects by specifying __module__ and __class__.

    Returns:
        The return value of `fn`
    """
    parameters = list(inspect.signature(fn).parameters.items())
    if inspect.isclass(fn):
        init_sig = getattr(fn, "__init__", None)
        if init_sig is not None:
            parameters = list(inspect.signature(init_sig).parameters.items())[1:]
    args = []
    kwargs = {}
    if isinstance(raw_data, dict):
        unused_args = raw_data.copy()
        for idx, (key, param) in enumerate(parameters):
            t = Any if param.annotation is inspect.Parameter.empty else param.annotation
            if param.kind in (
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.KEYWORD_ONLY,
            ):
                if param.default is inspect.Parameter.empty and key not in unused_args:
                    raise ValueError(f"Missing required argument {key!r} for {fn}")
                kwargs[key] = raw_to_typed(
                    unused_args.pop(key, param.default),
                    t,
                    strict,
                    allow_imports,
                    _path=key,
                    _stage=(idx,),
                )
            elif param.kind == inspect.Parameter.VAR_KEYWORD:
                for arg_key, arg_val in unused_args.items():
                    kwargs[arg_key] = raw_to_typed(
                        arg_val, t, strict, allow_imports, _path=key, _stage=(idx,)
                    )
                unused_args.clear()
            elif param.kind == inspect.Parameter.VAR_POSITIONAL:
                # No way to pass positional arguments
                pass
            elif param.kind == inspect.Parameter.POSITIONAL_ONLY:
                # No way to pass positional arguments
                raise RuntimeError(f"Unsupported positional only argument {key!r}")
            else:
                raise RuntimeError(f"Unknown parameter kind {param.kind!r}")
        if strict and len(unused_args) > 0:
            raise ValueError(f"Unexpected arguments: {unused_args!r}")
    elif isinstance(raw_data, list):
        unused_args = raw_data.copy()
        for idx, (key, param) in enumerate(parameters):
            t = Any if param.annotation is inspect.Parameter.empty else param.annotation
            if param.kind == inspect.Parameter.POSITIONAL_ONLY:
                if param.default is inspect.Parameter.empty and len(unused_args) == 0:
                    raise ValueError(
                        f"Missing required positional-only argument {key!r} at index {idx}"
                    )
                args.append(
                    raw_to_typed(
                        unused_args.pop(), t, strict, allow_imports, _path=key, _stage=(idx,)
                    )
                )
            elif param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                if param.default is inspect.Parameter.empty and len(unused_args) == 0:
                    raise ValueError(f"Missing required positional argument {key!r} at index {idx}")
                if len(unused_args) == 0:
                    arg_val = param.default
                else:
                    arg_val = unused_args.pop()
                args.append(
                    raw_to_typed(arg_val, t, strict, allow_imports, _path=key, _stage=(idx,))
                )
            elif param.kind == inspect.Parameter.VAR_POSITIONAL:
                for arg_val in unused_args:
                    args.append(
                        raw_to_typed(arg_val, t, strict, allow_imports, _path=key, _stage=(idx,))
                    )
                unused_args.clear()
            elif param.kind == inspect.Parameter.VAR_KEYWORD:
                # No way to pass keyword arguments
                pass
            elif param.kind == inspect.Parameter.KEYWORD_ONLY:
                raise RuntimeError(f"Unsupported keyword-only argument {key!r}")
            else:
                raise RuntimeError(f"Unknown parameter kind {param.kind!r}")
        if strict and len(unused_args) > 0:
            raise ValueError(f"Unexpected arguments: {unused_args!r}")
    else:
        raise ValueError(
            f"Cannot call function with raw data of type {type(raw_data)!r}, require list or dict"
        )
    return fn(*args, **kwargs)


float_pattern = re.compile(r"[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?")


def _split_dict_keys(dct: Dict[str, Any]) -> Dict[str, Any]:
    """Splits the given dict keys by first '.' to subdicts."""
    res = {}
    for key, value in dct.items():
        if "." in key:
            outer_key, _, inner_key = key.partition(".")
            if outer_key in res:
                if not isinstance(res[outer_key], dict):
                    raise ValueError(f"Cannot combine {outer_key!r} with {res!r}")
                res[outer_key][inner_key] = value
            else:
                res[outer_key] = {inner_key: value}
        else:
            if key in res:
                raise ValueError(f"Cannot combine {key!r} with {res!r}")
            res[key] = value

    return res


def _isinstance_deep(val: Any, tp_chk: Type) -> bool:
    """Verifies if the given value is an instance of the tp_chk, allowing for typing extensions."""
    if tp_chk is Any:
        return True
    elif typing.get_origin(tp_chk) is Literal:
        (value,) = typing.get_args(val)
        return val == value
    elif typing.get_origin(tp_chk) is list:
        (inner_type,) = typing.get_args(val)
        return isinstance(val, list) and all(_isinstance_deep(v, inner_type) for v in val)
    elif typing.get_origin(tp_chk) is tuple:
        inner_types = typing.get_args(val)
        if len(inner_types) == 2 and inner_types[1] == Ellipsis:
            return isinstance(val, tuple) and all(_isinstance_deep(v, inner_types[0]) for v in val)
        else:
            return (
                isinstance(val, tuple)
                and len(val) == len(inner_types)
                and all(_isinstance_deep(v, inner_type) for v, inner_type in zip(val, inner_types))
            )
    elif typing.get_origin(tp_chk) is dict:
        key_type, value_type = typing.get_args(val)
        return isinstance(val, dict) and all(
            _isinstance_deep(k, key_type) and _isinstance_deep(v, value_type)
            for k, v in val.items()
        )
    else:
        return isinstance(val, tp_chk)


def override(  # noqa: C901
    value: TType,
    overrides: Any,
    strict: bool = False,
    inst_type: Optional[Type[TType]] = None,
    allow_imports: bool = False,
    _path: str = "root",
    _stage: Tuple[int, ...] = (),
) -> TType:
    """
    Allows overriding values of a typed object using environment config.
    Allows overriding single config variables, or whole objects.

    Examples::

        class MyNamedTuple(NamedTuple):
            x: int
            y: str

        class MyNested(NamedTuple):
            nested: MyNamedTuple

        assert override(
            MyNested(nested=MyNamedTuple(x=42, y="foo")),
            {'nested.x': 5},
        ) == MyNested(nested=MyNamedTuple(x=5, y="foo"))
        assert override(
            MyNested(nested=MyNamedTuple(x=42, y="foo")),
            {'nested': '{"x": 5, "y": "bar"}'},
        ) == MyNested(nested=MyNamedTuple(x=5, y="bar"))

    Args:
        value: The base value to override.
        overrides: The overrides to apply
        strict: If true, no additional keys are allowed
        inst_type: If given, validate against this base type instead of the type of `value`.
        allow_imports: If true, allow instantiating types with dicts of __module__ and __class__.
        _path: Internal: The path to the current value.
        _stage: Internal: The current stage of the override.

    Returns:
        Same type as the input object (or `inst_type` if set), copied and updated from the
        overrides.
    """
    if inst_type is None:
        inst_type = type(value)
    type_name = getattr(inst_type, "__name__", repr(inst_type))
    if inst_type in (str, int, float, bool, None, type(None)):
        # Literal types
        if inst_type in (None, type(None)) and overrides == "None":
            overrides = None
        elif inst_type == bool and overrides in ("True", "true", "1", "False", "false", "0"):
            overrides = overrides in ("True", "true", "1")
        elif inst_type in (int, float) and isinstance(overrides, str):
            overrides = inst_type(overrides)
        if not isinstance(overrides, inst_type) and not (
            isinstance(overrides, int) and inst_type == float
        ):
            raise JsonValueError(
                f"Type does not match, expected {type_name} at {_path}, got {overrides!r}",
                inst_type,
                overrides,
                _path,
                _stage,
            )
        return overrides
    elif inst_type is Any:
        # Any
        if isinstance(overrides, str):
            if overrides.isnumeric():
                return int(overrides)
            elif overrides == "True":
                return True
            elif overrides == "False":
                return True
            return overrides
        if isinstance(value, (dict, list, tuple)):
            # Merge with dict, list, str
            return override(value, overrides, strict, type(value), allow_imports, _path, _stage)
        raise JsonValueError(
            f"Expected {type_name} at {_path}, got {overrides!r}",
            inst_type,
            overrides,
            _path,
            _stage,
        )
    elif typing.get_origin(inst_type) is Literal:
        # Literal[value]
        (value,) = typing.get_args(inst_type)
        if value != overrides:
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {overrides!r}",
                inst_type,
                overrides,
                _path,
                _stage,
            )
        return value
    elif typing.get_origin(inst_type) is Union:
        # Union[union_types[0], union_types[1], ...]
        union_types = typing.get_args(inst_type)
        if isinstance(overrides, str):
            for subtype in union_types:
                if subtype is None and overrides == "None":
                    return None
                elif subtype is bool:
                    if overrides == "True":
                        return True
                    elif overrides == "False":
                        return False
                elif subtype is int and overrides.strip().isnumeric():
                    return int(overrides)
                elif subtype is str:
                    return overrides
                elif subtype is float and float_pattern.fullmatch(overrides):
                    return float(overrides)
            if overrides.lstrip().startswith("{") or overrides.lstrip().startswith("["):
                overrides = json.loads(overrides)
            return raw_to_typed(
                overrides,
                inst_type,
                strict,
                allow_imports,
                _path,
                _stage,
            )
        for subtype in union_types:
            if _isinstance_deep(value, subtype):
                return override(
                    value,
                    overrides,
                    strict,
                    subtype,
                    allow_imports,
                    f"{_path} -> {getattr(subtype, '__name__', repr(subtype))}",
                    _stage + (1,),
                )
        raise JsonValueError(
            f"Expected {type_name} at {_path}, existing is {value!r} which is invalid",
            inst_type,
            value,
            _path,
            _stage,
        )
    elif (
        isinstance(inst_type, type)
        and issubclass(inst_type, tuple)
        and hasattr(inst_type, "__annotations__")
    ):
        # class MyClass(NamedTuple): ...
        if not isinstance(overrides, (dict, str)):
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {overrides!r}",
                inst_type,
                overrides,
                _path,
                _stage,
            )
        if isinstance(overrides, str):
            return raw_to_typed(
                json.loads(overrides),
                inst_type,
                strict,
                allow_imports,
                _path,
                _stage,
            )
        local_overrides = _split_dict_keys(overrides)
        if getattr(inst_type, "__dash_keys__", "False"):
            local_overrides = {key.replace("-", "_"): val for key, val in local_overrides.items()}
        kwargs = {
            field_name: (
                override(
                    getattr(value, field_name),
                    local_overrides.pop(field_name),
                    strict,
                    field_type,
                    allow_imports,
                    f"{_path} -> {type_name}:{field_name}",
                    _stage + (idx,),
                )
                if field_name in local_overrides
                else getattr(value, field_name)
            )
            for idx, (field_name, field_type) in enumerate(inst_type.__annotations__.items())
        }
        if strict and len(local_overrides) != 0:
            raise JsonValueError(
                f"Invalid config keys {', '.join(local_overrides.keys())} for {type_name} at "
                f"{_path}",
                inst_type,
                overrides,
                _path,
                _stage,
            )
        try:
            return inst_type(**kwargs)
        except BaseException:
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {overrides!r}",
                inst_type,
                overrides,
                _path,
                _stage,
            )
    elif dataclasses.is_dataclass(inst_type):
        # dataclass
        if not isinstance(overrides, (dict, str)):
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {overrides!r}",
                inst_type,
                overrides,
                _path,
                _stage,
            )
        if isinstance(overrides, str):
            return raw_to_typed(
                json.loads(overrides),
                inst_type,
                strict,
                allow_imports,
                _path,
                _stage,
            )
        local_overrides = _split_dict_keys(overrides)
        if getattr(inst_type, "__dash_keys__", "False"):
            local_overrides = {key.replace("-", "_"): val for key, val in local_overrides.items()}
        kwargs = {
            field.name: (
                override(
                    getattr(value, field.name),
                    local_overrides.pop(field.name),
                    strict,
                    field.type,
                    allow_imports,
                    f"{_path} -> {type_name}:{field.name}",
                    _stage + (idx,),
                )
                if field.name in local_overrides
                else getattr(value, field.name)
            )
            for idx, field in enumerate(dataclasses.fields(inst_type))
            if field.init
        }
        if strict and len(local_overrides) != 0:
            raise JsonValueError(
                f"Invalid config keys {', '.join(local_overrides.keys())} for {type_name} at "
                f"{_path}",
                inst_type,
                overrides,
                _path,
                _stage,
            )
        try:
            return inst_type(**kwargs)
        except BaseException:
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {overrides!r}",
                inst_type,
                overrides,
                _path,
                _stage,
            )
    elif (
        typing.get_origin(inst_type) is list
        or typing.get_origin(inst_type) is tuple
        or inst_type in (list, tuple)
    ):
        # List[inner_type] or Tuple[inner_type, Ellipsis] or
        # Tuple[inner_type[0], inner_type[1], ...]
        if inst_type == list:
            inner_type = Any
            inner_types = []
            cls = list
        elif inst_type == tuple:
            inner_type = Any
            inner_types = []
            cls = tuple
        elif typing.get_origin(inst_type) is list:
            (inner_type,) = typing.get_args(inst_type)
            inner_types = []
            cls = list
        else:
            inner_types = typing.get_args(inst_type)
            if len(inner_types) == 2 and inner_types[1] is Ellipsis:
                inner_type = inner_types[0]
            else:
                inner_type = None
            cls = tuple
        if not isinstance(overrides, (dict, str)):
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {overrides!r}",
                inst_type,
                overrides,
                _path,
                _stage,
            )
        if isinstance(overrides, str):
            return raw_to_typed(
                json.loads(overrides),
                inst_type,
                strict,
                allow_imports,
                _path,
                _stage,
            )
        local_overrides = _split_dict_keys(overrides)
        if not all(key.isnumeric() for key in local_overrides.keys()):
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {overrides!r}, expected integer keys",
                inst_type,
                overrides,
                _path,
                _stage,
            )
        local_overrides_int = {int(key): value for key, value in local_overrides.items()}
        new_max_idx = max(local_overrides_int.keys())
        original_max_idx = len(value)
        if inner_type is None and new_max_idx >= len(inner_types):
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {overrides!r}, index {new_max_idx} out of "
                f"bounds",
                inst_type,
                overrides,
                _path,
                _stage,
            )
        for i in range(original_max_idx, new_max_idx):
            if i not in local_overrides_int:
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {overrides!r}, missing value for index "
                    f"{i}",
                    inst_type,
                    overrides,
                    _path,
                    _stage,
                )
        return cls(
            (
                override(
                    value[idx],
                    local_overrides_int[idx],
                    strict,
                    inner_type,
                    allow_imports,
                    f"{_path} -> {idx}",
                    _stage + (idx,),
                )
                if idx in local_overrides_int
                else value[idx]
            )
            for idx in range(max(new_max_idx + 1, original_max_idx))
        )
    elif typing.get_origin(inst_type) is dict or inst_type == dict:
        # Dict[str, value_type]
        if inst_type == dict:
            value_type = Any
        else:
            key_type, value_type = typing.get_args(inst_type)
            assert key_type is str
        if not isinstance(overrides, (dict, str)):
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {overrides!r}",
                inst_type,
                overrides,
                _path,
                _stage,
            )
        if isinstance(overrides, str):
            return raw_to_typed(
                json.loads(overrides),
                inst_type,
                strict,
                allow_imports,
                _path,
                _stage,
            )
        local_overrides = _split_dict_keys(overrides)
        if getattr(inst_type, "__dash_keys__", "False"):
            local_overrides = {key.replace("-", "_"): val for key, val in local_overrides.items()}
        res = {
            key: (
                override(
                    subvalue,
                    local_overrides.pop(key),
                    strict,
                    value_type,
                    allow_imports,
                    f"{_path} -> {type_name}:{key!r}",
                    _stage + (idx,),
                )
                if key in local_overrides
                else subvalue
            )
            for idx, (key, subvalue) in value.items()
        }
        for key, val in local_overrides.items():
            if not isinstance(val, str):
                raise JsonValueError(
                    f"Expected new {type_name} at {_path} -> {type_name}:{key!r}, got {val!r}",
                    inst_type,
                    overrides,
                    _path,
                    _stage,
                )
            res[key] = raw_to_typed(
                json.loads(val),
                value_type,
                strict,
                allow_imports,
                f"{_path} -> {type_name}:{key!r}",
                _stage + (len(res),),
            )
        return res
    else:
        raise RuntimeError(f"Unknown type {inst_type}")


def to_json_object(obj: Any) -> Any:
    """
    Converts the given object to a json object.

    Args:
        obj: The object to convert

    Returns:
        The json-like object.
    """
    if isinstance(obj, (str, int, float, bool, type(None))):
        # Literal types
        return obj
    elif isinstance(obj, tuple) and hasattr(obj, "__annotations__"):
        # class MyClass(NamedTuple): ...
        return {
            field_name: to_json_object(getattr(obj, field_name))
            for field_name in obj.__annotations__.keys()
        }
    elif dataclasses.is_dataclass(obj):
        # dataclass
        return {
            field.name: to_json_object(getattr(obj, field.name))
            for field in dataclasses.fields(obj)
            if field.init
        }
    elif isinstance(obj, (list, tuple)):
        return [to_json_object(val) for val in obj]
    elif isinstance(obj, dict):
        return {key: to_json_object(val) for key, val in obj.items()}
    else:
        raise RuntimeError(f"Unknown type {type(obj)}")



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from megatron.energon.errors import SYSTEM_EXCEPTIONS, FatalSampleError
from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.worker import WorkerConfig, global_worker_config
from megatron.energon.wrappers._log_exception import log_exception
from megatron.energon.wrappers.base import BaseSingleWrapperDataset
from megatron.energon.wrappers.skip import SkipSample

T_batch = TypeVar("T_batch", covariant=True)
T_batch_sample = TypeVar("T_batch_sample", covariant=True)


class BatchDataset(
    BaseSingleWrapperDataset[T_batch_sample, T_batch], Generic[T_batch_sample, T_batch]
):
    """This dataset wrapper transforms a dataset of samples into a dataset of batches."""

    batch_size: int
    batcher: Callable[[List[T_batch_sample]], T_batch]
    drop_last: bool
    error_handler: Callable[[Exception, List[T_batch_sample]], None]
    worker_config: WorkerConfig

    def __init__(
        self,
        dataset: SavableDataset[T_batch_sample],
        batch_size: int,
        batcher: Callable[[List[T_batch_sample]], T_batch],
        drop_last: bool = False,
        error_handler: Callable[[Exception, List[T_batch_sample]], None] = log_exception,
        worker_config: Optional[WorkerConfig] = None,
    ):
        """Construct a BatchDataset.

        Args:
            dataset: The input dataset to wrap
            batch_size: The desired batch size. The last batch may be smaller.
            batcher: Function which combines separate samples into a single object. May raise
                :exc:`megatron.energon.SkipSample` to skip a sample.
            drop_last: If True, the last batch is dropped if it is smaller than the batch size.
            error_handler: Function which handles exceptions raised by the batcher. The default
                implementation logs the exception.
            worker_config: Configuration for the workers. Defaults to `global_worker_config`.
        """
        super().__init__(dataset)
        self.batch_size = batch_size
        self.batcher = batcher
        self.drop_last = drop_last
        self.error_handler = error_handler
        self.worker_config = worker_config or global_worker_config

    def __len__(self):
        n_samples = len(self.dataset)
        num_workers = max(self.worker_config.num_workers, 1)
        n_samples_per_worker_floor = n_samples // num_workers
        remaining_n_sample_workers = n_samples % num_workers
        n_batches_per_worker_floor = n_samples_per_worker_floor // self.batch_size
        if n_samples_per_worker_floor % self.batch_size != 0 and not self.drop_last:
            n_batches_per_worker_floor += 1
        # Correct number of batches for the workers which yield 1 more sample (to balance)
        n_batches_per_worker_ceil = (n_samples_per_worker_floor + 1) // self.batch_size
        if n_batches_per_worker_ceil % self.batch_size != 0 and not self.drop_last:
            n_batches_per_worker_ceil += 1

        return (
            n_batches_per_worker_floor * (num_workers - remaining_n_sample_workers)
            + n_batches_per_worker_ceil * remaining_n_sample_workers
        )

    def __iter__(self) -> Iterator[T_batch]:
        batch: List[T_batch_sample] = []
        for sample in self.dataset:
            batch.append(sample)
            if len(batch) == self.batch_size:
                try:
                    batch_sample = self.batcher(batch)
                    if isinstance(batch_sample, Generator):
                        yield from batch_sample
                    else:
                        yield batch_sample
                except SkipSample:
                    pass
                except SYSTEM_EXCEPTIONS:
                    raise FatalSampleError.from_sample(batch)
                except Exception as e:
                    self.error_handler(e, batch)
                batch = []
        if len(batch) > 0 and not self.drop_last:
            try:
                yield self.batcher(batch)
            except SkipSample:
                pass
            except SYSTEM_EXCEPTIONS:
                raise FatalSampleError.from_sample(batch)
            except Exception as e:
                self.error_handler(e, batch)

    def can_restore_sample(self) -> bool:
        return False

    def restore_sample(self, index: Tuple[Union[str, int], ...]) -> T_batch:
        # TODO: We'd need to store multiple indices to restore a batch
        # Also, returned elements don't support __restore_key__. Would need extension.
        raise NotImplementedError("BatchDataset does not support random access.")

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "batch_size": self.batch_size,
            "batcher": self._function_config(self.batcher),
            "drop_last": self.drop_last,
            "error_handler": self._function_config(self.error_handler),
            "worker_config": self.worker_config.config(),
            "dataset": self.dataset.config(),
        }

    def __str__(self):
        return f"BatchDataset(batch_size={self.batch_size}, drop_last={self.drop_last}, batcher={self.batcher}, dataset={self.dataset})"



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Any, Dict, Generic, Iterator, List, Optional, Tuple, TypeVar, Union

from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.worker import WorkerConfig, global_worker_config
from megatron.energon.wrappers.base import (
    BaseSingleWrapperDataset,
    BaseSingleWrapperMergedState,
    BaseSingleWrapperState,
)

T_sample = TypeVar("T_sample")


@dataclass
class RepeatState(BaseSingleWrapperState):
    offset: int


@dataclass
class RepeatMergedState(BaseSingleWrapperMergedState):
    offset: List[int]


class RepeatDataset(BaseSingleWrapperDataset[T_sample, T_sample], Generic[T_sample]):
    """This dataset repeats the inner dataset infinitely or a specific number of repeats."""

    repeats: Optional[int]
    worker_config: WorkerConfig
    _offset: List[int]

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        repeats: Optional[int] = None,
        worker_config: Optional[WorkerConfig] = None,
    ):
        """Construct a RepeatDataset.

        Args:
            dataset: The input dataset to repeat.
            repeats: Number of repeats, `None` for infinitely
            worker_config: Configuration for the workers. Defaults to `global_worker_config`.
        """
        super().__init__(dataset)
        self.repeats = repeats
        self.worker_config = worker_config or global_worker_config
        self._offset = [0] * max(self.worker_config.num_workers, 1)

    def __len__(self):
        if self.repeats is None:
            return len(self.dataset)
        return len(self.dataset) * self.repeats

    def __iter__(self) -> Iterator[T_sample]:
        worker_idx = self.worker_config.rank_worker_id()
        if self.repeats is None:
            while True:
                self._offset[worker_idx] += 1
                yield from self.dataset
                if self.worker_config.should_log(level=2):
                    self.worker_config.worker_log(
                        {
                            "t": "RepeatDataset.repeat",
                            "r": self.worker_config.rank,
                            "w": self.worker_config.rank_worker_id(),
                            "offset": self._offset[worker_idx],
                        }
                    )
        else:
            for offset in range(self._offset[worker_idx], self.repeats):
                self._offset[worker_idx] = offset + 1
                yield from self.dataset
                if self.worker_config.should_log(level=2):
                    self.worker_config.worker_log(
                        {
                            "t": "RepeatDataset.repeat",
                            "r": self.worker_config.rank,
                            "w": self.worker_config.rank_worker_id(),
                            "offset": self._offset[worker_idx],
                            "repeats": self.repeats,
                        }
                    )

    def save_state(self) -> RepeatState:
        return RepeatState.extend(
            super().save_state(),
            offset=self._offset[self.worker_config.rank_worker_id()],
        )

    def merge_states(self, states: List[RepeatState]) -> RepeatMergedState:
        assert all(s is None or isinstance(s, RepeatState) for s in states)
        return RepeatMergedState.extend(
            super().merge_states(states),
            offset=[0 if state is None else state.offset for state in states],
        )

    def restore_state(self, state: Optional[RepeatMergedState]) -> None:
        super().restore_state(state)
        if state is None:
            self._offset = [0] * max(self.worker_config.num_workers, 1)
        else:
            assert isinstance(state, RepeatMergedState)
            self._offset = state.offset

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
            "repeats": self.repeats,
            "worker_config": self.worker_config.config(),
        }

    def __str__(self):
        return f"RepeatDataset(repeats={self.repeats}, dataset={self.dataset})"
